import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import pandas as pd
import scipy.stats
import seaborn as sns
from functools import reduce
import os
import pickle

from ._clustering.clusterobj import clusterObj

def vis_all_clusters(clusterObj, x_label="time points", y_label="expression", Titles = "Cluster {col_name}", xtickslabels=None, **kwargs):
    """
    Visualize all the clusters with cluster prototype

    Parameters 
    -----------
    clusterObj
        input clustering object with results

    x_label
        x-axis label of the plot
    
    y_label
        y-axis label of the plot
    
    Titles :"Cluster {col_name}"
        titles for each cluster

    """
    if clusterObj.input_type =="expression":
        alltimeseries = np.array(clusterObj.DataSet.timeserieslist, dtype="double")
        tp = clusterObj.DataSet.timepts
    else:
        try:
            alltimeseries = np.array(clusterObj.DataSet.tiu_timeserieslist, dtype="double")
            tp = clusterObj.DataSet.tiu_timepts
        except ValueError:
            print("Please run iso.total_isoform_usage() to calculate total isoform usage before clustering.")

    median_allts = np.median(alltimeseries, axis=0)

    if clusterObj.transform:
        transfunc = clusterObj.TRANSFORMATION[clusterObj.transform] 
        median_allts = transfunc(median_allts)
        #if value returns infinite , set it as 0
        median_allts[~np.isfinite(median_allts)]=0
        
    

    ##faster probably
    reorder = []
    clusters = []
    for cluster, cluster_obj in clusterObj.index_clusters.items():
        for v in cluster_obj:
            reorder.append(v)
            clusters.append(cluster)

    precluster_tsarray = pd.DataFrame(median_allts)
    precluster_tsarray.columns = ["tp{}".format(x) for x in range(tp)]
    cluster_tsarray = precluster_tsarray.iloc[reorder,:].reset_index(drop=True)
    
    clusters_sns = pd.concat([pd.Series(clusters,name="clusters"),cluster_tsarray], axis=1)
    clusters_sns = pd.melt(clusters_sns, id_vars="clusters",value_vars=cluster_tsarray.columns, value_name="expression",var_name="timepoints", ignore_index=False)
    clusters_sns['gene'] = clusters_sns.index
    clusters_sns = clusters_sns.reset_index()

    #for prototype
    cluster_pro = pd.DataFrame(np.array(list(clusterObj._prototype.values())))
    cluster_pro.columns = ["tp{}".format(x) for x in range(tp)]
    cluster_pro['clusters'] = list(clusterObj.symbs_clusters.keys())
    cluster_pro = pd.melt(cluster_pro, id_vars='clusters', value_vars=cluster_pro.columns.to_list()[:-1], value_name="expression", var_name="timepoints")
 
    # allsns = pd.concat(cluster_sns_list)
    # sns.relplot(x="timepoints", y="expression",kind="line", hue="cluster", col="cluster",col_wrap=5,  height=3, aspect=.75, linewidth=2.5, palette= "Set2", data=allsns)

    grey = ['lightgray'] * len(clusterObj.genelist_clusters.keys())
    g = sns.relplot(data=clusters_sns, x="timepoints", y="expression", col="clusters", hue="clusters",
    units="gene", kind="line", height=3, aspect=.75,estimator=None, linewidth=1, palette=grey, legend=False, **kwargs)

    g.set_axis_labels(x_label, y_label, fontsize=13)
    g.set_titles(Titles)
    if xtickslabels:
        g.set_xticklabels(labels=xtickslabels,rotation=90, ha="right", fontsize=8)

    ##plot prototypes
    #g = sns.relplot(data=cluster_pro, x="timepoints", y="expression", kind="line", col="clusters",hue="clusters",col_wrap=3, height=3, aspect=.75, linewidth=4, palette="Set2")
    pal = sns.color_palette("dark", len(clusterObj.genelist_clusters.keys()))
    i=0
    for x,ax in g.axes_dict.items():
        subdata = cluster_pro[cluster_pro['clusters']==x]
        
        sns.lineplot(data=subdata, x="timepoints", y="expression", linewidth=4, color=pal[i], ax=ax, legend=False)
        i+=1

    plt.show()



def vis_cluster(clusterObj, cluster):
    alltimeseries = np.array(clusterObj.DataSet.timeserieslist, dtype="double")
    mean_allts = np.mean(alltimeseries, axis=0)

    cluster_obj = clusterObj.index_clusters[cluster]

    cluster_tsarray = pd.DataFrame(mean_allts[cluster_obj,:])
    cluster_tsarray.columns = ["tp{}".format(x) for x in range(clusterObj.DataSet.timepts)]

    cluster_sns = pd.melt(cluster_tsarray, value_vars=cluster_tsarray.columns, value_name="expression", var_name="timepoints")
    cluster_sns['cluster'] = "Cluster {}".format(cluster)

    sns.relplot(x="timepoints", y="expression",kind="line", col_wrap=5,  height=3, aspect=.75, linewidth=2.5, palette= "Set2", data=cluster_sns)

def cluster_heatmap(clusterObj, scalelab="z-score", xlab="Total isoform usage across time", xticklabels =["E17-PN1", "PN1-PN10", "PN10-28","PN28-PN90"],  z_score=None, standard_scale=None, cmap="coolwarm", save_path=None, cluster_label_space = 0.0001):
    """
    Visualize all clusters in a heatmap.

    Parameters
    -----------
    clusterObj
        clusterobj
    
    scalelab
        label for the scale
    
    xlab
        label for x-axis

    xticklabels
        label for x-ticks
    
    z_score : (default:None, 0, 1)
        visualize z_score value rowwise(0) or columnwise(1), cannot be used together with standard_scale

    standard_scale : (default:None, 0, 1)
        visualize standard_scale value rowwise(0) or columnwise(1), cannot be used together with z_score

    cmap : "coolwarm"
        colormap
    
    save_path
        file path for saving the figure.
    
    Return
    -------
    Figure
    
    """
    reorder=[]
    clusters=[]
    for u,v in clusterObj.index_clusters.items():
        for i in v:
            clusters.append(u)
            reorder.append(i)
    pal = sns.color_palette("Set2", 100)
    rowc = dict(zip(list(clusterObj.index_clusters.keys()), pal))
    rowcol = pd.Series(clusters, name='Clusters').map(rowc)
    rowcol.index = pd.DataFrame(clusterObj.DataSet.timeserieslist[0]).iloc[reorder,:].index

    if cmap != "coolwarm":
        colmap = sns.light_palette(cmap, as_cmap=True)

    else:
        colmap = cmap
    cl = sns.clustermap(pd.DataFrame(clusterObj.DataSet.timeserieslist[0]).iloc[reorder,:], standard_scale=standard_scale, z_score=z_score, cmap=colmap, 
                row_cluster=False, 
                col_cluster=False,
                yticklabels=False, row_colors=rowcol, xticklabels=xticklabels)


    cl.ax_heatmap.set(xlabel=xlab)

    ratios=0
    for u,v in clusterObj.index_clusters.items():

        ratios += len(v)*cluster_label_space
        cl.ax_row_dendrogram.annotate("Cluster "+str(u),(0.5,1-ratios))
    plt.xlabel(scalelab)
    plt.show()
    
    if save_path:
        plt.savefig(save_path)

def vis_heatmap(list_genes, dataset, scalelab="z-score", xlab="Total isoform usage across time", xticklabels =None,  z_score=None, standard_scale=None, cmap="coolwarm", save_path=None, cluster_label_space = 0.0001):
    """
    Visualize all clusters in a heatmap.

    Parameters
    -----------
    list_genes
        list of genes ID you want to visualize the expression value
    
    dataset
        a dataset object

    scalelab
        label for the scale
    
    xlab
        label for x-axis

    xticklabels
        label for x-ticks
    
    z_score : (default:None, 0, 1)
        visualize z_score value rowwise(0) or columnwise(1), cannot be used together with standard_scale

    standard_scale : (default:None, 0, 1)
        visualize standard_scale value rowwise(0) or columnwise(1), cannot be used together with z_score

    cmap : "coolwarm"
        colormap
    
    save_path
        file path for saving the figure.
    
    Return
    -------
    Figure
    
    """
    reorder=[]
    clusters=[]
    for u,v in clusterObj.index_clusters.items():
        for i in v:
            clusters.append(u)
            reorder.append(i)
    pal = sns.color_palette("Set2", 100)
    rowc = dict(zip(list(clusterObj.index_clusters.keys()), pal))
    rowcol = pd.Series(clusters, name='Clusters').map(rowc)
    rowcol.index = pd.DataFrame(clusterObj.DataSet.timeserieslist[0]).iloc[reorder,:].index

    targetset = dataset.timeserieslist[:,np.where(np.isin(dataset.gene_id, list_genes))[0],:]
    if cmap != "coolwarm":
        colmap = sns.light_palette(cmap, as_cmap=True)
    else:
        colmap = cmap
        
    cl = sns.clustermap(pd.DataFrame(dataset.timeserieslist[0]).iloc[reorder,:], standard_scale=standard_scale, z_score=z_score, cmap=colmap, 
                row_cluster=False, 
                col_cluster=False,
                yticklabels=False, row_colors=rowcol, xticklabels=xticklabels)


    cl.ax_heatmap.set(xlabel=xlab)

    ratios=0
    for u,v in clusterObj.index_clusters.items():

        ratios += len(v)*cluster_label_space
        cl.ax_row_dendrogram.annotate("Cluster "+str(u),(0.5,1-ratios))
    plt.xlabel(scalelab)
    plt.show()
    
    if save_path:
        plt.savefig(save_path)

def compare_mod(mods , groups, group_type="clusters", figw=3, top=None, figh=None):
        def heatmap(x, y, size, color, figw, figh):
            fig,ax= plt.subplots(figsize=(figw,figh))

            # Mapping from column names to integer coordinates
            x_labels = [v for v in x.unique()]
            y_labels = [v for v in y.unique()]
            x_to_num = {p[1]:p[0]+0.5 for p in enumerate(x_labels)} 
            y_to_num = {p[1]:p[0]+0.5 for p in enumerate(y_labels)} 

            size_scale = 50
            ax.grid()
            #ax.hlines(y=y.map(y_to_num), xmin=0, xmax=x.map(x_to_num), color='black',linewidth=3, alpha =0.8)
            tpc=ax.scatter(
                x=x.map(x_to_num), # Use mapping for x
                y=y.map(y_to_num), # Use mapping for y
                s=size, # Vector of square sizes, proportional to size parameter
                marker='o',# Use square as scatterplot marker
                c=color,
                cmap='cividis'
            )
            cbar = fig.colorbar(tpc)

            # Show column labels on the axes
            ax.set_xticks([x_to_num[v] for v in x_labels])
            ax.set_xticklabels(x_labels, rotation=45, horizontalalignment='right')
            ax.set_yticks([y_to_num[v] for v in y_labels])
            ax.set_yticklabels(y_labels)
            cbar.set_label("-log10(adjusted p-values)")
            plt.xlim([0,len(x.unique())])


        df = pd.DataFrame()

        for e,i in enumerate(mods):
            if isinstance(i, dict):
                for uu,vv in i.items():
                    vv[0]=pd.DataFrame(vv[0])
                    vv[0]['group'] = f"cluster{groups[e]}_module{uu}"
                    vv[0] = vv[0].sort_values(['Adjusted P-value'])
                    if top is not None:
                        df =pd.concat([df, vv[0].iloc[:top,:]])
                    else:
                        df =pd.concat([df, vv[0]])
            else:
                
                if len(i)>0:
                    i[0]['group'] = f"{group_type}{groups[e]}"
                    i[0]=pd.DataFrame(i[0])
                    i[0] = i[0].sort_values(['Adjusted P-value'])
                    if top is not None:
                            df =pd.concat([df, i[0].iloc[:top,:]])
                    else:
                        df =pd.concat([df, i[0]])
            
            #mod_d['group'] = 'TSIS iso'
        if df.shape[0] == 0:
            return None
        else:
            df = df[df['Adjusted P-value']<0.05].sort_values(['group','Adjusted P-value'], ascending=False)
            # if df.shape[0]>20:
            #     df = df.iloc[:21,:]

            heatmap(x=df['group'], y=df['Term'], color=-np.log10(df['Adjusted P-value']), size=-np.log10(df['Adjusted P-value'])*50, figh=df.shape[0]/3, figw=figw)

                
            return df


def switch_plot(gene, DataSet, ascov, xaxis_label=None, all_isoforms=False, relative_abundance=False):
    """
    Switching plot for isoforms / Expression plot for non-switched genes
    if the input gene is not isoform switched, the expression plot will be plotted.
    
    Parameters 
    ----------
    gene : str
        Input gene ID / symbs you would like to plot
    DataSet : DataSet obj
    ascov : DataFrame
        the result dataframe of your isoform switch detection
    xaxis_label : list
        x axis label for the plots
    """
    try:
        entrezid = DataSet.gene_id[np.where(DataSet.symbs==gene)[0]]
    except:
        entrezid = DataSet.gene_id[np.where(DataSet.gene_id==gene)[0]]

    if str(entrezid[0]) in set(ascov['gene']) and all_isoforms==False:
        ncol = 4
        nrow =  2 if len(entrezid) > 4 else len(entrezid) % 4
    
        #fig, axes = plt.subplots(nrows=nrow, ncols=ncol)
        
        for axx, idx in enumerate(np.where(ascov['gene']==str(entrezid[0]))[0]):
            
            iso1 = ascov['major_transcript'].to_list()[idx]
            iso2 = ascov['minor_transcript'].to_list()[idx]

            
            iso1_id = np.where(DataSet.transcript_id == iso1)[0]
            iso2_id = np.where(DataSet.transcript_id == iso2)[0]
            
            
            if relative_abundance:
                val_name="relative abundance"
                iso_dict = DataSet.isoobj.normdf
                try:
                    iso1_ts = pd.DataFrame(iso_dict[entrezid[0]]['normarr'][:,np.where(iso_dict[entrezid[0]]['indices']==iso1_id)[0][0],:])
                    iso2_ts = pd.DataFrame(iso_dict[entrezid[0]]['normarr'][:,np.where(iso_dict[entrezid[0]]['indices']==iso2_id)[0][0],:])
                except:
                    iso1_ts = pd.DataFrame(np.squeeze(iso_dict[entrezid[0]]['normarr'][:,np.where(iso_dict[entrezid[0]]['indices']==iso1_id)[0][0],:], axis=1))
                    iso2_ts = pd.DataFrame(np.squeeze(iso_dict[entrezid[0]]['normarr'][:,np.where(iso_dict[entrezid[0]]['indices']==iso2_id)[0][0],:], axis=1))
            else:
                val_name="expression"
                try:
                    iso1_ts = pd.DataFrame(np.squeeze(DataSet.timeserieslist[:,iso1_id,:], axis=1))
                    iso2_ts = pd.DataFrame(np.squeeze(DataSet.timeserieslist[:,iso2_id,:], axis=1))
                except:
                    iso1_ts = pd.DataFrame(DataSet.timeserieslist[:,iso1_id,:])
                    iso2_ts = pd.DataFrame(DataSet.timeserieslist[:,iso2_id,:])

            iso1_ts = pd.melt(iso1_ts, value_vars=iso1_ts.columns, value_name=val_name, var_name="time point")
            iso2_ts = pd.melt(iso2_ts, value_vars=iso2_ts.columns, value_name=val_name, var_name="time point")
            iso1_ts['isoform'] = iso1
            iso2_ts['isoform'] = iso2

            isos = pd.concat([iso1_ts, iso2_ts], axis=0)

            g=sns.catplot(x="time point", y=val_name, hue="isoform", capsize=.2, kind="point", data=isos)
            plt.title(gene, fontsize=15)
            g.set_xticklabels(labels=xaxis_label, ha="right", fontsize=12)
            sns.catplot(x="time point", y=val_name, hue="isoform", kind="strip", data=isos)
            g.set_xticklabels(labels=xaxis_label, ha="right", fontsize=12)
            plt.title(gene, fontsize=15)
    
    else:
        
        entrezids = np.where(DataSet.gene_id==gene)[0]
        if len(entrezids) == 0:
            entrezids = np.where(DataSet.symbs==gene)[0]

        allts = []
        if relative_abundance:
            iso_dict = DataSet.isoobj.normdf
            val_name = "relative abundance"
            try:
                iso_ts = pd.DataFrame(np.squeeze(iso_dict[entrezid[0]]['normarr'], axis=1))
            except:
                iso_ts = iso_dict[entrezid[0]]['normarr']

            for eachts in range(iso_ts.shape[1]):
                tmp_ts = pd.DataFrame(iso_ts[:,eachts,:])
                tmp_ts = pd.melt(tmp_ts, value_vars=tmp_ts.columns, value_name=val_name, var_name="time point")
                tmp_ts['isoform'] = DataSet.transcript_id[iso_dict[entrezid[0]]['indices'][eachts]]
                allts.append(tmp_ts)
        else: 
            val_name="expression"
            for idx in entrezids:
                try:
                    iso_ts = pd.DataFrame(np.squeeze(DataSet.timeserieslist[:,idx,:], axis=1))
                except:
                    iso_ts = pd.DataFrame(DataSet.timeserieslist[:,idx,:])
                    
                iso_ts = pd.melt(iso_ts, value_vars=iso_ts.columns, value_name=val_name, var_name="time point")
                iso_ts['isoform'] = DataSet.transcript_id[idx]
                allts.append(iso_ts)

        isos = pd.concat(allts, axis=0)
        g=sns.catplot(x="time point", y=val_name, hue="isoform", capsize=.2, kind="point", data=isos)
  
        g.set_xticklabels(labels=xaxis_label, ha="right", fontsize=12)
        plt.title(gene, fontsize=15)

def expression_plot(gene, DataSet):
    entrezids = np.where(DataSet.gene_id==gene)[0]
    
    if len(entrezids) == 0:
        entrezids = np.where(DataSet.symbs==gene)[0]

    allts = []
    for idx in entrezids:
        try:
            iso_ts = pd.DataFrame(np.squeeze(DataSet.timeserieslist[:,idx,:], axis=1))
        except:
            iso_ts = pd.DataFrame(DataSet.timeserieslist[:,idx,:])
            
        iso_ts = pd.melt(iso_ts, value_vars=iso_ts.columns, value_name="expression", var_name="time point")
        allts.append(iso_ts)

    isos = pd.concat(allts, axis=0)
    # sumisos = isos.groupby("time point").agg(sum)
    # sumisos = sumisos.reset_index()

    g= sns.catplot(x="time point", y= "expression", kind="point",data=isos)
    plt.title(gene, fontsize=15)



def gsea_plot(gsea_result, cluster, modules=None, nterms=None):
    """
    Visualizing the functional enrichment

    Parameters
    -----------
    gsea_result : dict 
        the results of gsea
    cluster : str
        the cluster number you would like to visualize
    nterms : int 
        (optional) if you would like to visualize only subset of terms e.g. the top 10 terms 
    """
    cluster = int(cluster)

    if modules is None:
        mod_cluster = gsea_result[cluster]
    else:
        mod_cluster = gsea_result[cluster][modules]

    if isinstance(mod_cluster, dict):
        subset = pd.DataFrame(mod_cluster[0])
    else: 
        if mod_cluster[0].shape[0]>0:
            subset = pd.DataFrame(mod_cluster[0])
            subset = subset.sort_values(['Adjusted P-value'])

        else:
            print("This gsea has no result.")

    if nterms:
        subset = subset.head(nterms)

    plt.figure(figsize=(subset.shape[0],12))
    subset['-log10(adjusted p-value)'] = -np.log10(subset['Adjusted P-value'])
    sns.barplot(y="Term", x="-log10(adjusted p-value)", data=subset, color="salmon")
    if modules is not None:
        plt.title(f"Cluster {cluster} Module {modules}")
    else:
        plt.title(f"Cluster {cluster}")


def vis_modules(mods, dataset, cluster, size=5, outputpng=None):
    ascov = dataset.isoobj.is_result

    mapping = dict(zip(dataset.gene_id,dataset.symbs))

    def _map_color(x):
        xx = str(x).split("/")[0]
        if xx in set(ascov['gene'].tolist()):
            try:
                domains = np.unique(reduce(lambda x,y: x+y, ascov[ascov['gene_symb']==xx]['exclusive_domains'].tolist()))
            except:
                domains = []

            if len(domains)>0:
                return "#e06666" #pink IS with functional switch
            else:
                return "#fff2cc" ##yellow without functional switch
        else:
            return "#b4f7fe" ##green #interactors

    def _map_entrez2symb(x):
        xx=x.split("/")
        if xx[0] in mapping.keys():
            symb = mapping[xx[0]]
        else: 
            symb = x
        if len(xx)>1:
            return symb+"/"+xx[1]
        else:
            return symb
    
    
    v = mods[cluster]
    
    nmod = np.sum([True if len(eachmod.nodes) > size else False for eachmod in v[0]])#check the how many modules bigger than the threshold size
    nrow = (nmod//3)+1 if nmod/3 - nmod//3 > 0 else nmod//3
    h=nrow*5
    if nrow==0:
        nrow=1
        h=10
    elif nrow==1:
        nrow=2
        h=nrow*5
    
    fig, ax=plt.subplots(nrow, 3, figsize=(h,h))
    if len(ax.shape)==2:
        for e,G in enumerate(zip(v[0], [x for vv in ax for x in vv])):
            if len(G[0].nodes)>size:
                colors = list(map(_map_color, G[0].nodes))
                nx.draw(G[0], with_labels=True, labels = dict(zip(G[0].nodes, list(map(_map_entrez2symb, G[0].nodes)))), node_color=colors, node_size=300, font_size=10, ax=G[1])
            # nt.from_nx(G)
            # nt.show('nx.html')
                G[1].set_title(f"cluster {cluster} module {e} : q-val: {v[1][e]}")
            
        if outputpng:
            plt.savefig(f"{outputpng}/modules_clu{cluster}.png", dpi=200, bbox_inches="tight")
            plt.close()
        else:
            plt.axis('off')
            plt.show()
        

    else:
        for e, G in enumerate(zip(v[0], [x for x in ax])):
            colors = list(map(_map_color, G[0].nodes))
            nx.draw(G[0], with_labels=True, labels = dict(zip(G[0].nodes, list(map(_map_entrez2symb, G[0].nodes)))), node_color=colors, node_size=300, font_size=10, ax=G[1])
            G[1].set_title(f"cluster {cluster} module {e}")

        if outputpng:
            plt.savefig(f"{outputpng}/modules_clu{cluster}.png", dpi=200, bbox_inches="tight")
            plt.close()
        else:
            plt.axis('off')
            plt.show()

def _make_dot(edges, mapping, ascov, html, related_genes):
    isgene = ascov['gene'].unique()
    nxnet = nx.Graph(list(edges.edges))
    nxdot = nx.drawing.nx_pydot.to_pydot(nxnet)
    infile = open(os.path.join("./spycone_pkg/spycone/data/network/network_human_PPIDDI.tab.pkl"),'rb')
    jointgraph = pickle.load(infile)
    def map_entrez2symb(x):
        if int(x) in mapping.keys():
            symb = mapping[int(x)]
        else:
            symb = x
        return symb
    mapper = list(map(map_entrez2symb, nxnet.nodes))
    
    allnodes = [nxdot.get_nodes()[x].get_name() for x in range(len(nxdot.get_nodes()))]
    for e,n in enumerate(allnodes):
        nxdot.get_node(n)[0].obj_dict['attributes']['label'] = mapper[e]
        if mapper[e] in set(related_genes):
            nxdot.get_node(n)[0].obj_dict['attributes']['color']="#01aeb2"
            nxdot.get_node(n)[0].obj_dict['attributes']['penwidth']=5
        nxdot.get_node(n)[0].obj_dict['attributes']['style'] = "filled"
        nxdot.get_node(n)[0].obj_dict['attributes']['fontname'] = "verdana"
        if str(n) in set(isgene):
            nxdot.get_node(n)[0].obj_dict['attributes']['fillcolor'] = "#CFA3EA"
        else:
           nxdot.get_node(n)[0].obj_dict['attributes']['fillcolor'] = "#FBA27B"
    
    ##check if interaction is affected
    alledges = [nxdot.get_edges()[x].obj_dict['points'] for x in range(len(nxdot.get_edges()))]
    for e, edge in enumerate(alledges):
        no1 = edge[0]
        no2 = edge[1]
        domain1 = ascov[ascov['gene']==no1][['exclusive_domains']].explode('exclusive_domains').exclusive_domains.unique()
        domain2 = ascov[ascov['gene']==no2][['exclusive_domains']].explode('exclusive_domains').exclusive_domains.unique()

        style=""
        if domain1.shape[0]>0:
            for do in domain1:
                try:
                    checknode = str(no1)+"/"+do
                    for jointedge in jointgraph.edges:
                        if jointedge[0] == checknode and str(no2) in jointedge[1]:
                            style="dashed"
                        if jointedge[1] == checknode and str(no2) in jointedge[0]:
                            style="dashed"
                except:
                    continue
        
        if domain2.shape[0]>0:
            for do in domain2:
                try:
                    checknode = str(no2)+"/"+do
                    for jointedge in jointgraph.edges:
                        if jointedge[0] == checknode and str(no1) in jointedge[1]:
                            style="dashed"
                        if jointedge[1] == checknode and str(no1) in jointedge[0]:
                            style="dashed"
                except:
                    continue

        nxdot.get_edges()[e].obj_dict['attributes']['style'] = style
        nxdot.get_edges()[e].obj_dict['attributes']['penwidth'] = 3
        
    return(nxdot)
    # nt = Network(height="1000px", width="1000px", notebook=True)
    # nt.from_nx(nxnet)
    # nt.options.edges.color.inherit = False
    # nt.options.groups = """{
    #     2: {color:{background:'green'}},
    #     1: {color:{background:'yellow'}},
    #     0: {color:{background:'lightblue'}}
    # }"""
    # nt.show_buttons()
    # nt.show(html)


def vis_better_modules(dataset, mod, cluster, dir, related_genes=set(), module=None):
    """
    Visualize modules with pydot in SVG format.

    Parameters 
    ------------
    dataset : 
        dataset obj
    mod : 
        DOMINO results
    cluster :
        Cluster number to visualize
    dir :
        Local directory to save the images
    related_genes : (Optional)
        Set of genes to change the color of the nodes (color of the node outer border)
    """
    ascov1 = dataset.isoobj.is_result 
    #ascov1['gene'] = list(map(str, list(map(lambda x: int(x) if not np.isnan(x) else x, ascov1['gene']))))
    mapping = pickle.load(open("/nfs/proj/spycone/spycone_pkg/spycone/data/entrez2symb.pkl", "rb"))

    if module:
        edges = mod[cluster][0][module]
        nxdot = _make_dot(edges, mapping, ascov1, f"{dir}modules/c{cluster}m{module}.html", related_genes)
        nx.drawing.nx_pydot.to_pydot(nxdot,  f"{dir}/modules/c{cluster}m{module}.dot")
    else:
        mods = mod[cluster][0]
        for e,v in enumerate(mods):
            if len(v.nodes)>4:
                nxdot=_make_dot(v, mapping, ascov1, f"{dir}modules/c{cluster}m{e}.html", related_genes)
                nxdot.obj_dict['attributes']['label'] = f"Cluster {cluster} Module {e+1}: \n adjusted p-val: {mod[cluster][1][e]:.2E}"
                nxdot.obj_dict['attributes']['fontsize'] = 17
                nxdot.obj_dict['attributes']['fontname'] = "verdana"
                nxdot.write_svg(f"{dir}/modules/c{cluster}m{e+1}.svg")

        




import networkx as nx
import pandas as pd


def build_network(network_file):
    """"""
    if type(network_file)!=str:
        return network_file.networkz
    else:
        edges_dataset = pd.read_csv(network_file, sep='\t', header=None, dtype=str)
        edges = []
        for ind, row in edges_dataset.iterrows():
            # if row.iloc[0]!=row.iloc[2]:
            edges.append((row.iloc[0], row.iloc[1]))
        G = nx.Graph()
        G.add_edges_from(edges)
        nx.set_node_attributes(G, 0, 'score')

        return G



import networkx as nx
import pandas as pd
import numpy as np
from joblib import Parallel, delayed
import itertools as it

from ..DataSet import DataSet
from ..BioNetwork import BioNetwork

#compute pairwise network location similarity
#input as gene node names, corresponding to time series index and network node names


# def inverse_shortest_path(gene_names, BioNetwork):
#     #total node count - 1
#     G = BioNetwork.g()
#     maxdist = len(BioNetwork.lst_g())-1
#     #dist_map = G.new_vertex_property("double", np.inf)
#     mapper = {}
#     for i in range(0, G.num_vertices()):
#         mapper.setdefault(G.vertex_properties['name'][i],i)


#     #initiate
#     #get unique gene list 
#     gene_index = [mapper[x] for x in gene_names]
#     sp_distmat = np.zeros((G.num_vertices(), G.num_vertices()))
    
#     #
    

#     tmp_distmat = shortest_distance(G)


#     #get result from property map
#     for i in range(G.num_vertices()):
#         sp_distmat[i,:] = tmp_distmat[i].a
    

#     #graph tool return a large number when no path is found...
#     #correct those large number so no negative return
#     sp_distmat[sp_distmat>maxdist] = maxdist
#     return (maxdist-sp_distmat)/maxdist, gene_index





import pandas as pd
import numpy as np
import time
import random
import warnings
import logging
from collections import defaultdict
from sklearn.cluster import *
from sklearn_extra.cluster import KMedoids
from sklearn.metrics.cluster import adjusted_rand_score

from ..DataSet import DataSet
from ..BioNetwork import BioNetwork
#from similarity.calculate_similarities import calculate_similarities_timeseriesobj_to_medoid, calculate_similarities_clusterobject, calculate_center_medoid
from .._prototype.prototype_builder import _prototype_builder
from ....archive.qc import _qc_silhouette
from ..clustering import clustering
from .build_cluster import _add_obj_to_most_sim_cluster, cluster_map_to_item, merge_above_similarity_threshold_clusters
from .clusterobj import clusterObj
from . import similarities as sim

class clara(clusterObj):
    def __init__(self, DataSet, n_cluster, n_start, metric="euclidean",prototypefunction = "median", n_samples=5,
                samplesize=None, genelist_clusters=None, index_clusters=None, symbs_clusters=None, _prototype=None, _final_n_cluster=None):
        self.DataSet= DataSet
        self.n_cluster = n_cluster
        self.n_start = n_start
        self.metric = metric
        self.prototypefunction=prototypefunction
        self.n_samples = n_samples
        self.samplesize = samplesize
        super().__init__(genelist_clusters, index_clusters, symbs_clusters, _prototype, _final_n_cluster)

        def _recommended_samplesize(self):
                n = self.DataSet.ts[0].shape[0]
                k = self.n_cluster

                if n <= 250:
                    return n
                
                return 40+2*k # original ticone is 40+2*k
            
        if samplesize == None:
            self.samplesize = _recommended_samplesize(self)

    def _add_clusters(self, genelistclusters, symbsclusters, indexclusters):
        self.genelist_clusters=genelistclusters
        self.symbs_clusters=symbsclusters
        self.index_clusters=indexclusters

        return None

    def fit(self):
        #return clusterings 

        #start iteration
        bestSim = -np.inf
        bestclusters = defaultdict(list)
        i=0

        while i < self.n_samples:
            #select random samples
            sampledata = self.DataSet.timeserieslist.copy() #which is 3D
            print(sampledata.shape)

            timeseriesobj_ids = np.arange(sampledata.shape[1])
            chosen_index = random.sample(set(timeseriesobj_ids), self.samplesize)

            #fit to k medoids
            sampledata_median = np.median(sampledata[:,chosen_index,:], axis=0)

            print(sampledata_median.shape)

            thisclustering = KMedoids(self.n_cluster)
            thisclustering_fit = thisclustering.fit(sampledata_median)

            #predict non samples
            non_sample_id = np.delete(timeseriesobj_ids, chosen_index)
            nonsampledata = np.median(sampledata[:,non_sample_id,:], axis=0)

            predict_sample = thisclustering.predict(nonsampledata)

            
            #calculate scores
            allindex = np.hstack([chosen_index, non_sample_id])
            alllabels = np.hstack([thisclustering_fit.labels_,predict_sample])
            print(len(chosen_index), non_sample_id.shape)
            print(thisclustering_fit, predict_sample.shape)

            assert allindex.shape[0] == alllabels.shape[0]

            x = np.transpose(np.vstack([alllabels, allindex]))
            
            index_clusters = dict()
            for k,v in x:
                if k in index_clusters.keys():
                    index_clusters[k].append(v)
                else:
                    index_clusters.setdefault(k,[v])
            
            #medoids dict
            thismedoids = dict()
            for u in thisclustering.medoid_indices_:
                thismedoids.setdefault(u,self.DataSet.timeserieslist[:,u,:])


            totalsim = thisclustering.inertia_
            print(totalsim)

            if bestSim <= totalsim:
                bestSim = totalsim
                bestclusters = index_clusters

            i+=1

        genelist_clusters, symbs_clusters = cluster_map_to_item(self.DataSet, bestclusters)
        cluster_prototypes = _prototype_builder(self.DataSet.timeserieslist, bestclusters, self.prototypefunction)
        self._prototype = cluster_prototypes
        self._add_clusters(genelist_clusters, symbs_clusters, bestclusters)
        self._final_n_cluster = len(bestclusters)

        return bestclusters


from collections import defaultdict
import numpy as np
from Bio import motifs
from Bio.Seq import Seq
import pandas as pd
import os
import pickle
import requests
import seaborn as sns
import matplotlib.pyplot as plt
from functools import reduce
from scipy.stats import pearsonr
from scipy.stats import mannwhitneyu, fisher_exact, kruskal
from Bio.SeqUtils import GC
from joblib import Parallel, delayed
import gc


from ._util_stat.multipletesting import pvalue_correction

dir_path = os.path.dirname(os.path.realpath(__file__))

class SF_coexpression():
    """
    Return the coexpression between Splicing factor and isoforms
    
    Parameters 
    ------------
    dataset : input dataset object

    padj_method : Multiple testing method. Default=Bonferonni

    corr_cutoff :  Correlation coefficient cutoff. Dafault=0.7

    padj_cutoff : Adjusted p-value cutoff. Default=0.05

    method : Method to calculate the correlation value.
    
    Return
    -------
    Create an instance for co-expression analysis of splicing factors and transcript abundance.

    """
    def __init__(self, dataset, padj_method="bonf", corr_cutoff=0.7, padj_cutoff = 0.05, method="pearson"):
        self.dataset=dataset
        self.padj_method = padj_method
        self.corr_cutoff = corr_cutoff
        self.method= method
        

    def _get_splicing_factors(self):
        sfs=[]
        with open("./spycone_pkg/spycone/data/network/splicingfactors.txt", "r") as f:
            tmp=f.readlines()
            for ss in tmp:
                sfs.append(ss.replace("\n", ""))

        sfs = pd.DataFrame(sfs, columns=['GeneSymbol'])
        mapper = dict(zip(self.dataset.symbs, self.dataset.gene_id))
        sfs['entrezid'] = list(map(lambda x: mapper[x] if x in self.dataset.symbs else None, sfs['GeneSymbol']))
        sfs = sfs[sfs['GeneSymbol'].isin(self.dataset.symbs)]
        print(str(np.sum(sfs['GeneSymbol'].isin(self.dataset.symbs)))+" of the splicing factor are in the dataset")
    
        return sfs

    def _get_matrices(self, sfs, timelag=1):
        ##calculate isoform abundances
        #
        isoabun = np.empty(shape=(self.dataset.shape[0], 0, self.dataset.shape[2]))
        transcript_idx = []
        geneid = []
        for gene, info in self.dataset.isoobj.normdf.items():
            if len(info['indices'])>1:
                transcript_idx.append(info['indices'])
                isoabun = np.concatenate((isoabun, info['normarr']), axis=1)
                for _ in range(info['normarr'].shape[1]):
                    geneid.append(gene)
        transcript_idx = reduce(lambda x,y: x+y, transcript_idx)
        transcript = self.dataset.transcript_id[transcript_idx]
        isoabun = np.concatenate(isoabun, axis=1) ##isoform abundance of isoforms shape = (isoforms, time points*replicates)

        #correlate splicing factor with isoform abundance
        ###splicing factor expression shape = (splicing factor, time points*replicates)
        sfs_expression = np.empty(shape=(self.dataset.shape[0], 0, self.dataset.shape[2]))
        sf_id = []
        for e,sf in enumerate(sfs['entrezid']):
            tmp = self.dataset.isoobj.normdf[sf]
            sfexp = np.expand_dims(np.sum(tmp['array'], axis=1),axis=1)
            sfs_expression=np.concatenate([sfs_expression, sfexp], axis=1)
            sf_id.append(sfs['GeneSymbol'].tolist()[e])

        sfs_expression = np.concatenate(sfs_expression, axis=1)

        ##filter for only IS genes
        isgenes_idx = np.where(np.isin(geneid, self.dataset.isoobj.is_result['gene']))[0]
        final_isoabun = isoabun[isgenes_idx,:]
        final_transcript = transcript[isgenes_idx]
        final_geneid = [geneid[x] for x in isgenes_idx]


        ##filter splicing factors variances
        highvar = np.where(np.var(sfs_expression, axis=1)>1)[0]
        final_sfsexp =sfs_expression[highvar,:] 
        final_sfid = [sf_id[x] for x in highvar]

        ##
        final_mat = np.concatenate([final_isoabun, final_sfsexp])
        mat_index = np.append(final_transcript, final_sfid)

        assert final_mat.shape[0] == mat_index.shape[0]
        return final_mat, mat_index, final_transcript, final_geneid, final_sfid

    def _corrcoef_loop(self, matrix):
        rows, cols = matrix.shape[0], matrix.shape[1]
        r = np.ones(shape=(rows, rows))
        p = np.ones(shape=(rows, rows))
        for i in range(rows):
            for j in range(i+1, rows):
                if self.method=="pearson":
                    r_, p_ = pearsonr(matrix[i], matrix[j])
                if self.method=="softdtw":
                    pass
                r[i, j] = r[j, i] = r_
                p[i, j] = p[j, i] = p_
        return r, p


    def _calculate_coexpression(self, final_mat, final_transcript, final_geneid, final_sfid):
        ##correlation with p values #######tmp read the results below
        sf_iso_mat, sf_iso_pval = self._corrcoef_loop(final_mat)
        
        ####multiple correction
        mapper = dict(zip(self.dataset.gene_id, self.dataset.symbs))

        sf_iso_07 = sf_iso_mat[final_transcript.shape[0]+1:, 0:final_transcript.shape[0]]
        sf_iso_dict = {"sf":[], "iso":[], "iso_gene":[], "iso_genesymb":[], "corr":[], "pval":[]}
        for sf in range(sf_iso_07.shape[0]):
            for iso in range(sf_iso_07.shape[0]):
                sf_iso_dict['sf'].append(final_sfid[sf])
                sf_iso_dict['iso'].append(final_transcript[iso])
                sf_iso_dict['iso_gene'].append(final_geneid[iso])
                sf_iso_dict['iso_genesymb'].append(mapper[final_geneid[iso]])
                sf_iso_dict['corr'].append(sf_iso_07[sf,iso])
                sf_iso_dict['pval'].append(sf_iso_pval[sf][iso])


        ##correct for multiple testing
        sf_iso_df = pd.DataFrame(sf_iso_dict)
        sf_iso_df = sf_iso_df.sort_values("pval")
        mp = pvalue_correction(sf_iso_df['pval'], method=self.padj_method)
        sf_iso_df['padj']=mp.corrected_pvals
        
        ##cleaning data, 
        sf_iso_df['cluster'] = list(map(self._map_to_clusters, sf_iso_df['iso_gene']))
        sf_iso_df = sf_iso_df[(np.abs(sf_iso_df['corr'])>self.corr_cutoff) & (sf_iso_df['padj']<0.05)].sort_values("iso_gene")
        sf_iso_df['bin_corr'] = [1 if x > 0 else -1 for x in sf_iso_df['corr']]

        sf_group = sf_iso_df.groupby(["bin_corr","iso_gene","iso_genesymb", "cluster"]).aggregate(lambda x: tuple(x))
        if sf_group.shape[0] == 0:
            return "No significant correlated SF is found."

        everycluster = pd.DataFrame(sf_group.groupby(["cluster", "bin_corr"]).aggregate(lambda x: tuple(x))['sf'].apply(lambda x: np.unique(reduce(lambda a,y: a+y, x)))).reset_index()
        #union_sf : the sf that are only found in one direction

        

        ####get list of SF that is positively and negatively correlated to at least one of the isoform in the cluster
        
        everycluster['diff_sf']=everycluster.groupby("cluster").aggregate(lambda x: list(x))['sf'].apply(lambda x: np.repeat(set(x[0]).symmetric_difference(set(x[1])),2) if len(x)>1 else [x[0]]).explode('sf')

        unique_sf = []
        union_sf =[]
        for x,s in everycluster.iterrows():
            unique_sf.append(set(s['sf']).intersection(set(s['diff_sf'])))
            union_sf.append(set(s['sf']).difference(set(s['diff_sf'])))

        everycluster['unique_sf']=unique_sf
        everycluster['union_sf']=union_sf
        return sf_iso_df, everycluster

    def _map_to_clusters(self, gene):
        #sf_iso_clu = []
        #for gene in sf_iso_df['iso_gene']:
        match=False
        for u,v in self.dataset.clusterobj.genelist_clusters.items():
            if gene in set(v):
                match=True
                break
        if match:
            return u
        #sf_iso_df['cluster'] = sf_iso_clu


    def coexpression_with_sf(self):
        sfs = self._get_splicing_factors()
        final_mat, mat_index, list_transcript, list_geneid, list_sfid = self._get_matrices(sfs)
        sf_iso_df, cluster_sf = self._calculate_coexpression(final_mat, list_transcript, list_geneid, list_sfid)

        return sf_iso_df, cluster_sf

def create_motifsobj():
        ########this only read the consensus
        # motifs_df = pd.read_csv(os.path.join(dir, "sfanalysis/motif_info.csv"), sep="\t")
        # print(motifs_df.head())
        # motifs_df['motifs']=motifs_df['motifs'].apply(eval)
        # instances_pre = motifs_df[motifs_df['motifname']==self.list_SF[0]]['motifs'].to_list()

        # motif_obj = []
        # for instance in instances_pre:
        #     instance = list(map(lambda x: x.replace("N",""), instance))
        #     print(instance)
        #     instances = list(map(Seq, instance))
        #     m = motifs.create(instances)

        #     motif_obj.append(m)
        dir = "/nfs/proj/spycone/"
        # return motif_obj
        motif_obj = defaultdict(list)
        for mat in os.listdir(os.path.join(dir, "sfanalysis/eCLIP_PWM")):
            if ".txt" in mat :
                with open(os.path.join(dir, "sfanalysis/eCLIP_PWM",mat)) as handle:
                    record = motifs.parse(handle, "TRANSFAC")
                motif_obj[mat.split(".")[1]].append(record)

        with open(os.path.join(dir_path, "data/network/splicingfactors.txt"),"w") as f:
            for u in motif_obj.keys():
                f.write(u+"\n")
            f.close()

        # with open(os.path.join(dir,"spycone_pkg/spycone/data/motif_obj.pkl"), "wb") as f:
        #     pickle.dump(motif_obj, f, pickle.HIGHEST_PROTOCOL)
        
        # f = open(os.path.join(dir,"spycone_pkg/spycone/data/motif_obj.pkl"), "rb")
        # motif_obj = pickle.load(f)
        motif_pssm = defaultdict(list)
        for sf, motif_o in motif_obj.items():
            for e,bpmotif in enumerate(motif_o):
                if len(bpmotif)==1:          
                    
                    pwm=bpmotif[0].counts.normalize()
                    # background = {"A":(1-gc_ratio)/2,"C":gc_ratio/2,"G":gc_ratio/2,"T":(1-gc_ratio)/2}
                    # pssm=pwm.log_odds(background)
                    motif_pssm[sf].append([pwm])

        with open(os.path.join(dir,"spycone_pkg/spycone/data/motif_pwm.pkl"), "wb") as f:
            pickle.dump(motif_pssm, f, pickle.HIGHEST_PROTOCOL)

        return motif_obj, motif_pssm


class SF_motifsearch():
    '''
    Return the PSSM score of target SF and exons binding.

    Parameters
    ------------
    list_SF : list
        List of splicing factors / RBPs. E.g. the SF that is co-expressed to the isoform abundance. 

    list_genes : list
        List of genes you want to check for SF binding sites. E.g. the cluster of genes that the input SF is co-expressed. 

    gtf_df : str or dataframe
        gtf file path / dataframe

    gc_ratio : GC content that makes up the background for PSSM score calculation. Default=0.6.

    flanking : Flanking region size 

    Return
    -------
    Create an instance for SF motif enrichment analysis.

    '''
    def __init__(self, list_SF, list_genes, dataset, gtf, gc_ratio=0.6, flanking=400):
        self.list_SF = list_SF
        self.list_genes = list_genes
        self.dataset = dataset
        self.gtf = gtf
        self.flanking = flanking
        #self.motif_pssm = pickle.load(open(os.path.join(dir_path,"data/motif_pssm.pkl"), "rb"))
        self.motif_obj = pickle.load(open(os.path.join(dir_path,"data/motif_obj.pkl"), "rb"))
        self.motif_pwm = pickle.load(open(os.path.join(dir_path,"data/motif_pwm.pkl"), "rb"))
        self.motif_thres = pickle.load(open(os.path.join(dir_path,"data/motif_thres.pkl"), "rb"))

    ###perform motifs search
    def get_seq(self, exid, site="5p", bg=False):
        target = self.gtf[self.gtf['exon_id']==exid]
        if target.shape[0]!=0:
            targetpos=target[['seqname','start','end','exon_id']]
            
            targetpos = targetpos.set_index(['exon_id'])
            targetpos = targetpos[['seqname', 'start','end']].astype(str)

            # if bg:
            #     exon_len = int(targetpos['end'].values[0])-int(targetpos['start'].values[0])+1
            #     random_start = np.random.randint(1,45000000, size=1)
            #     url = f"https://api.genome.ucsc.edu/getData/sequence?genome=hg38;chrom=chr{targetpos['seqname'].values[0]};start={str(int(random_start)-self.flanking)};end={str(int(random_start)+exon_len+self.flanking)}"
            #     response = requests.get(url)
            # else:
            if site=="5p":
                url = f"https://api.genome.ucsc.edu/getData/sequence?genome=hg38;chrom=chr{targetpos['seqname'].values[0]};start={str(int(targetpos['start'].values[0])-self.flanking)};end={str(int(targetpos['start'].values[0])+self.flanking)}"
            
            elif site=="3p":
                url = f"https://api.genome.ucsc.edu/getData/sequence?genome=hg38;chrom=chr{targetpos['seqname'].values[0]};start={str(int(targetpos['end'].values[0])-self.flanking)};end={str(int(targetpos['end'].values[0])+self.flanking)}"

            response = requests.get(url)
            
            if 'dna' in response.json().keys():
                return response.json()['dna']
            else:
                return None

    def _exon_search(self, exons, sf, site):
        sf_results=defaultdict(list)
        motifs_pwm = self.motif_pwm[sf]
        for exon in exons:
            exon_seq = self.get_seq(exid=exon, site=site)
            if exon_seq is None:
                continue
            else:
                for e,pwm in enumerate(motifs_pwm):  
                    if len(pwm)==1:          
                        test_seq = Seq(exon_seq)
                        submotstr = f'{sf}_{e}'
                        gc_ratio = round(GC(test_seq)/100,2)
      
                        background = {"A":(1-gc_ratio)/2,"C":gc_ratio/2,"G":gc_ratio/2,"T":(1-gc_ratio)/2}
                        pssm=pwm[0].log_odds(background)

                        # distribution = pssm[0].distribution(background=background, precision=10**4)
                        # threshold = distribution.threshold_patser()
                        submotifs = None
                        submotstr = None
                        # for pos, sc in pssm[0].search(test_seq, self.threshold):
                        #     dire = np.sign(pos)

                        # #relpos_start = np.abs(pos) - self.flanking
                        # #relpos_end = np.abs(pos) - len(exon_seq) + self.flanking
                        #     submotifs = f"{exon};{dire};{pos};{sc}"
                        #     submotstr = f'{sf}_{e}'
                        
                        submotstr = f'{sf}_{e}'
                        allscores = pssm.calculate(test_seq)
                        # if self.motif_thres[submotstr] is not None:
                        #     allscores = pssm[0].search(test_seq, threshold=self.motif_thres[submotstr])
                            
                        # for pos, sc in enumerate(allscores):
                        #     submotifs = f"{exon};1;{pos};{sc}"
                        #     sf_results[submotstr].append(submotifs)
                        #### this only save the scores not the position  
                        this_threshold = self.motif_thres[submotstr] 
                        sf_results[submotstr] = [sc if not np.isnan(sc) else next if sc > this_threshold else next for sc in allscores]
                    else:
                        continue
            gc.collect()
        return sf_results

    def _parallel_search(self, sf, exon_dicts, site, bg=False):
        sf_results = []
        motif_str = []
        #for each SF search in every exons
        #for gene, exons in self.target_exons_dict.items():
        # with concurrent.futures.ProcessPoolExecutor() as executor:
        #     for gene, each_gene_res in zip(exon_dicts.values(), executor.map(self._exon_search, exon_dicts.values(), np.repeat(sf, len(exon_dicts.values())))):
        #         sf_results.append(each_gene_res)
      
        sf_results = Parallel()(delayed(self._exon_search)(exons=v, sf=sf, site=site) for v in exon_dicts.values())
        # for v in exon_dicts.values():
        #     res = self._exon_search(v, sf, site=site)
        #     sf_results.append(res)
  
        sf_per_results = defaultdict(list)
        for eachgene in sf_results:
            for eachmot, eachres in eachgene.items():
                if eachmot is not None:
                    for x in eachres:
                        sf_per_results[eachmot].append(x)
        return sf_per_results
            
    def search_motifs(self, exons="loss", site="5p"):
        """
        Calculate PSSM scores on all exons in the input list of genes. 

        Parameters
        -----------
        

        """
        #get exons from IS results
        target_exons = self.dataset.isoobj.is_result[self.dataset.isoobj.is_result['gene_symb'].isin(self.list_genes)]
        target_exons_dict = defaultdict(list)
        n_exons =0
        if exons=="loss":
            for x in target_exons.iterrows():
                target_exons_dict[x[1]['gene_symb']]=x[1]['loss_exons']
                n_exons+=len(x[1]['loss_exons'])
        else:
            for x in target_exons.iterrows():
                target_exons_dict[x[1]['gene_symb']]=x[1]['gain_exons']
                n_exons+=len(x[1]['gain_exons'])

        self.target_exons_dict = target_exons_dict

        
        sf_results = defaultdict(list)
        motif_str =defaultdict(list)
        #get motif objects from list SF
        # motif_obj, motif_pssm = self.create_motifsobj()
        # self.motif_obj = motif_obj
        # self.motif_pssm = motif_pssm
        # parallelize motif searching 
        ## return a list of result in the following format:
         ### exonID;direction of the strand;matched position of the seq;score from biopython;relative pos to the start site (including flanking region);relative pos to the end + flanking region; length of the exon; 
        for sf in self.list_SF:
            each_sf_res = self._parallel_search(sf, target_exons_dict, site=site)
            sf_results[sf].append(each_sf_res)
        # iterate the results to create a dictionary where keys are SF and values are the results for each exon
        ### note that each SF might have multiple motifs, in this case the motifs are denoted as SF_1, SF_2 from the same SF.
        self.motif_result = sf_results
        gc.collect()
        return sf_results, n_exons

    def search_background_motifs(self, site="5p"):
        allexons = self.gtf["exon_id"].to_numpy()
        #get exons from IS results
        if "self.target_exons_dict" not in locals():
            ValueError("Please run search_motifs first.")

        bg_exons_dict=defaultdict(list)
        n_exons=0
        target_exons = self.dataset.isoobj.is_result[self.dataset.isoobj.is_result['gene_symb'].isin(self.list_genes)]
        for x in target_exons.iterrows():
             bg_exons_dict[x[1]['gene_symb']]=x[1]['constant_exons']
             n_exons+=len(x[1]['constant_exons'])

        self.bg_exons_dict =  bg_exons_dict

        print("get exons bg dict")
        #get motif objects from list SF
        sf_results = defaultdict(list)
        motif_str =defaultdict(list)
        #get motif objects from list SF
        # motif_obj, motif_pssm = self.create_motifsobj()
        # self.motif_obj = motif_obj
        # self.motif_pssm = motif_pssm
        # parallelize motif searching 
        ## return a list of result in the following format:
         ### exonID;direction of the strand;matched position of the seq;score from biopython;relative pos to the start site (including flanking region);relative pos to the end + flanking region; length of the exon; 
        
       # with concurrent.futures.ProcessPoolExecutor() as executor:
        #    for each_sf_res, sf in zip(self.list_SF, executor.map(self._parallel_search, self.list_SF)):
        for sf in self.list_SF:
            each_sf_res = self._parallel_search(sf, bg_exons_dict, site=site)
            sf_results[sf].append(each_sf_res)
        # iterate the results to create a dictionary where keys are SF and values are the results for each exon
        ### note that each SF might have multiple motifs, in this case the motifs are denoted as SF_1, SF_2 from the same SF.

        self.motif_background = sf_results
        gc.collect()
        return sf_results, n_exons
    

    def _df_forvis(self, lossres, gainres, bg=None):
        '''
        MannWhitney U test or Fisher exact test

        '''
        def _checkscore(x):
            if np.isfinite(x):
                return x
            else:
                return 0

        vis_sf = []
        ldf = {#'allpos':[],
        'logscores':[],
        #'allscores':[],
        'sf_motif':[],
        'threshold':[]}
        for sf,v in lossres[0].items():
            for eachmot, exons in v[0].items():
                for ex in exons:
                    # signs = np.sign(float(ex.split(";")[2]))
                    # signss = 1 if signs == 0 else signs
                    # signsss = "+ strand" if signss==1 else "- strand"
                    #ldf['allpos'].append(np.abs(float(ex.split(";")[2]))-self.flanking)
                    thisscore=float(ex)
                    ldf['logscores'].append(thisscore)
                    # df['alldir'].append(signsss)
                    ldf['sf_motif'].append(eachmot)
                    ldf['threshold'].append(self.motif_thres[eachmot])

        ldf = pd.DataFrame(ldf)
        ldf['rawtype'] = "lost exons"
        ldf['type'] = f"LE ({lossres[1]})"
        vis_sf.append(ldf)

        gdf = {#'allpos':[],
        'logscores':[],
        #'allscores':[],
        'sf_motif':[],
        'threshold':[]}
        #countexons=[]
        for sf,v in gainres[0].items():
            for eachmot, exons in v[0].items():
                for ex in exons:
                    #countexons.append(ex)
                    # signs = np.sign(float(ex.split(";")[2]))
                    # signss = 1 if signs == 0 else signs
                    # signsss = "+ strand" if signss==1 else "- strand"
                    #gdf['allpos'].append(np.abs(float(ex.split(";")[2]))-self.flanking)
                    thisscore=float(ex)
                    gdf['logscores'].append(thisscore)
                    #gdf['allscores'].append(10**thisscore)
                    # df['alldir'].append(signsss)
                    gdf['sf_motif'].append(eachmot)
                    gdf['threshold'].append(self.motif_thres[eachmot])

        gdf = pd.DataFrame(gdf)
        gdf['rawtype'] = "gained exons"
        gdf['type'] = f"GE ({gainres[1]})"
        
        vis_sf.append(gdf)

        if bg:
            bgdf = {#'allpos':[],
            'logscores':[],
        #'allscores':[],
        'sf_motif':[],
        'threshold':[]}
            for sf,v in bg[0].items():
                for eachmot, exons in v[0].items():
                    for ex in exons:
                        #countexons.append(ex.split(";")[0])
                        # signs = np.sign(float(ex.split(";")[2]))
                        # signss = 1 if signs == 0 else signs
                        # signsss = "+ strand" if signss==1 else "- strand"
                        #bgdf['allpos'].append(np.abs(float(ex.split(";")[2]))-self.flanking)
                        thisscore=float(ex)
                        bgdf['logscores'].append(thisscore)
                        #bgdf['allscores'].append(10**thisscore)
                        # bgdf['alldir'].append(signsss)
                        bgdf['sf_motif'].append(eachmot)
                        bgdf['threshold'].append(self.motif_thres[eachmot])

            bgdf = pd.DataFrame(bgdf)
            bgdf['rawtype'] = "background"
            bgdf['type'] = f"UE ({bg[1]})"
            vis_sf.append(bgdf)
        
        alldf = pd.concat(vis_sf)
        alldf = alldf[alldf['threshold'].notnull()]
        alldf=alldf[(np.isfinite(alldf['logscores']))].reset_index()
        #alldf['titles'] = alldf['sf_motif']+", "+alldf['alldir']
        return alldf

    def _get_pvals(self, alldf, method="mannwhitney"):
        alldf = alldf[alldf['logscores']>alldf['threshold']]
        ### cal pvalue
        if method=="mannwhitney":
            pvals =defaultdict(list)
            #for stra in ["- strand", "+ strand"]:
                #ddf = df[df['alldir']==stra].reset_index()
                #bgddf = bgdf[bgdf['alldir']==stra].reset_index()
            for sf in alldf['sf_motif'].unique():
                idx=alldf[(alldf['sf_motif']==sf) & (alldf['rawtype']=="gained exons")].index
                lidx=alldf[(alldf['sf_motif']==sf) & (alldf['rawtype']=="lost exons")].index
                bgidx=alldf[(alldf['sf_motif']==sf) & (alldf['rawtype']=="background")].index
                gexons = [alldf['logscores'][x] for x in idx]
                lexons = [alldf['logscores'][x] for x in lidx]
                bgs = [alldf['logscores'][x] for x in bgidx]
                
                try:
                    pvals[sf].append(mannwhitneyu(lexons, bgs, alternative="greater")[1]) #[lost, gain]
                except ValueError:
                    pvals[sf].append(1)

                try:
                    pvals[sf].append(mannwhitneyu(gexons, bgs, alternative="greater")[1]) #[lost, gain]
                except ValueError:
                    pvals[sf].append(1)

                try:
                    pvals[sf].append(mannwhitneyu(lexons, gexons, alternative="greater")[1]) #[lost, gain]
                except ValueError:
                    pvals[sf].append(1)


        return pvals

    def vis_motif_density(self, res=None, style="2D", bg=None, method="mannwhitney", pval_cutoff=None, merge_motif=False, alldf=None, pvals=None, palette=None, sf_list=None):
        if alldf is None and pvals is None:
            if res is None:
                raise("Please input result from search_motif() as res")
            alldf, pvals = self._df_pval_forvis(res, bg=bg, method=method)
        

        filtered_alldf = alldf[alldf['logscores']>alldf['threshold']]
        
        def print_pvalues(g, i, j, hdiff, pvid):
            for x,ax in enumerate(g.axes_dict.values()):
                text=pvals[ax.get_title()][pvid]

                px = []
                ph=[]
                phmin = []
                for p in alldf['type'].unique():
                    tmp = filtered_alldf[(filtered_alldf['type']==p)&(filtered_alldf['sf_motif']==ax.get_title())]['logscores']
                    if not np.isnan(tmp.max()):
                        ph.append(tmp.max()) ##max of each group
                        phmin.append(filtered_alldf[(filtered_alldf['type']==p)&(filtered_alldf['sf_motif']==ax.get_title())]['logscores'].min())
                    else:
                        ph.append(0)
                        phmin.append(0)

                pxx = (i+j)/2
                phh = max(ph[i],ph[j])
                
                props = {'connectionstyle':'bar','arrowstyle':'-',
                            'shrinkA':20,'shrinkB':20,'linewidth':2}
                ax.plot([i,i,j,j], [phh+hdiff, phh+hdiff+0.2, phh+hdiff+0.2, phh+hdiff], lw=1.5, c="k")
                ax.text(s=f"p={str('{:.2e}'.format(text))}", x=pxx-0.3, y=phh+hdiff+0.3)
                ax.set_ylim((np.min(phmin)-1, phh+5))

        if sf_list:
            filtered_alldf = filtered_alldf[filtered_alldf['sf_motif'].isin(sf_list)]
            
        if pval_cutoff:
            sigs_sf = []
            for u,v in pvals.items():
                checksig = False
                for vv in v:
                    if vv<=pval_cutoff:
                        checksig=True

                if checksig == True:
                    sigs_sf.append(u)

            
            sigs_df = filtered_alldf[filtered_alldf['sf_motif'].isin(sigs_sf)]

            palette1 = defaultdict(str)
            for e,x in enumerate(sigs_df['type'].unique()):
                palette1[x] = list(palette.values())[e]
                
            if style=="boxplot":
                plt.figure(figsize=[8,9])
                g= sns.catplot(data=sigs_df, x="type",  y="logscores",  col="sf_motif", col_wrap=5, kind="box",linewidth=3, palette=palette1, sharey=False)
                g.set_titles("{col_name}")
                print_pvalues(g, 0, 1, 0.5, 2) #loss vs gain
                print_pvalues(g, 1, 2, 1, 1) #gain vs bg
                print_pvalues(g, 0, 2, 2, 0) #loss vs bg
                # for x,ax in enumerate(g.axes_dict.values()):
                #     thistitle = ax.get_title()
                #     pvvv = pvals[thistitle][3]
                #     ax.set_title(thistitle+"\n"+f"kruskal p-value: {str('{:.2e}'.format(pvvv))}", y=0.95)
                g.set_axis_labels(x_var="Type", y_var="log(Score)")
                g.set_xticklabels()

        else:
            if style=="boxplot":
                palette1 = defaultdict(str)
                for e,x in enumerate(filtered_alldf['type'].unique()):
                    palette1[x] = list(palette.values())[e]

                plt.figure(figsize=[8,9])
                g= sns.catplot(data=filtered_alldf, x="type",  y="logscores",  col="sf_motif", col_wrap=5, kind="box",linewidth=3, palette=palette1, sharey=False)
                g.set_titles("{col_name}")
                print_pvalues(g, 0, 1, 0.5, 2) #loss vs gain
                print_pvalues(g, 1, 2, 1, 1) #gain vs bg
                print_pvalues(g, 0, 2, 1.5, 0) #loss vs bg
                # for x,ax in enumerate(g.axes_dict.values()):
                #     thistitle = ax.get_title()
                #     pvvv = pvals[thistitle][3]
                #     ax.set_title(thistitle+"\n"+f"kruskal p-value: {str('{:.2e}'.format(pvvv))}", y=0.95)
                g.set_axis_labels(x_var="Type", y_var="log(Score)")
                g.set_xticklabels()
            
        



# """
# Created on Fri Sep  6 10:37:15 2019

# @author: alexander
# """
from collections import defaultdict
from logging import raiseExceptions

import pandas as pd
import numpy as np
import os, sys
import gseapy as gp
import time
import warnings


sys.path.insert(0, os.path.abspath('./_NEASE/nease/'))
from ._NEASE.nease import nease

# Disable
def _blockPrint():
    sys.stdout = open(os.devnull, 'w')

# Restore
def _enablePrint():
    sys.stdout = sys.__stdout__




def list_gsea(genelist, taxid, gene_sets=None, is_results=None, cutoff=0.05, method="gseapy"):
    """
    Perform gene set enrichment on a list of gene
    """
    geneset_map = {10090:['KEGG_2019_Mouse'], 9606:['KEGG_2019_Human']}
    org_map = {10090:'Mouse', 9606:'Human'}
    bg = {10090:'mmusculus_gene_ensembl', 9606:'hsapiens_gene_ensembl'}

    if method == "gseapy":
        if gene_sets is None:
            gene_sets = geneset_map[taxid]
        
        enr = gp.enrichr(gene_list = genelist,
                        gene_sets = gene_sets,
                        organism=org_map[taxid],
                        cutoff=cutoff,
                        background=bg[taxid]
                        )
        
        enr_results = enr.results[enr.results['Adjusted P-value']<cutoff]
        time.sleep(2)

        print("---------Gene Set Enrichment Result---------\n")
        print(f"Method: {method} Database: {gene_sets}")
        print(f"{len(genelist)} of genes found enriched in {enr_results.shape[0]} terms.")
        print("-----END-----")
        
        return enr_results


def clusters_gsea(DataSet, taxid, gene_sets=None, is_results=None, cutoff=0.05, method="nease"):
    """
    Perform gene set enrichment on clusters (cluster object)

    """
    
    geneset_map = {10090:['KEGG_2019_Mouse'], 9606:['KEGG_2019_Human']}
    org_map = {10090:'Mouse', 9606:'Human'}
    bg = {10090:'mmusculus_gene_ensembl', 9606:'hsapiens_gene_ensembl'}
    X = DataSet.clusterobj
    
    warnings.simplefilter("ignore")
    if method == "gseapy":
        if gene_sets is None:
            gene_sets = geneset_map[taxid]
        

        enr_results = defaultdict(list)
        for u,v in X.symbs_clusters.items():
            enr = gp.enrichr(gene_list = v,
                            gene_sets = gene_sets,
                            organism=org_map[taxid],
                            cutoff=cutoff,
                            background=bg[taxid]
                            )
            
            enr_results[u].append(enr.results[enr.results['Adjusted P-value']<cutoff])
            time.sleep(2)

        print("---------Gene Set Enrichment Result---------\n")
        print(f"Method: {method} Database: {gene_sets}")
        for u,v in enr_results.items():
            print("Cluster {}".format(u)," found enriched in {} terms.".format(v[0].shape[0]))
        print("-----END-----")
        
        return enr_results, None

    if method == "nease":
        nease_support = ['PharmGKB','HumanCyc','Wikipathways','Reactome','KEGG','SMPDB','Signalink','NetPath','EHMN','INOH','BioCarta','PID']

        ##default database == "KEGG"
        if gene_sets is None:
            gene_sets = ["KEGG"]

        check_support_database = [True if geneset not in nease_support else False for geneset in gene_sets]
        if np.sum(check_support_database) > 0:
            raise ValueError("Please input a supported database: ['PharmGKB','HumanCyc','Wikipathways','Reactome','KEGG','SMPDB','Signalink','NetPath','EHMN','INOH','BioCarta','PID']")

        if is_results is None:
            raise ValueError("Please provide the result dataframe of isoform switch detection for nease enrichment.")

        _blockPrint()
        enr_results = defaultdict(list)
        nease_obj = defaultdict(list)
        for u,v in X.genelist_clusters.items():
            subset = is_results[is_results['gene'].astype(str).isin(list(map(lambda x: str(int(x)), v)))]
            nease_input = []
            for eachrow in subset.iterrows():
                domains = eachrow[1]['exclusive_domains']
                if len(domains)>0:
                    for domain in domains:
                        nease_input.append(str(eachrow[1]['gene'])+"/"+domain)

            if len(nease_input)>0:
                nease_input = pd.DataFrame({"domain":nease_input})
                events=nease.run(nease_input, organism="Human", input_type="Spycone")
                nease_enr=events.enrich(database=gene_sets)

                if nease_enr is not None:
                    nease_enr=nease_enr.rename(columns = {'adj p_value':'Adjusted P-value'})
                    nease_enr=nease_enr.rename(columns = {'Pathway name':'Term'})
                    nease_obj[u].append(events)
                    enr_results[u].append(nease_enr[nease_enr['Adjusted P-value']<cutoff])
                else:
                    continue
            else:
                continue
        
        _enablePrint()
        print("---------Gene Set Enrichment Result---------\n")
        print(f"Method: {method} Database: {gene_sets}")
        for u,v in enr_results.items():
            print("Cluster {}".format(u)," found enriched in {} terms.".format(v[0].shape[0]))
        print("-----END-----")
        
        return enr_results, nease_obj

def modules_gsea(X, clu, taxid, type="PPI", gene_sets=None, cutoff=0.05, method="nease"):
    """
    Perform gene set enrichment on network modules after domino
    """
    geneset_map = {10090:['KEGG_2019_Mouse'], 9606:['KEGG_2019_Human']}
    mapping = dict(zip(clu.DataSet.gene_id, clu.DataSet.symbs))
    org_map = {10090:'Mouse', 9606:'Human'}
    bg = {10090:'mmusculus_gene_ensembl', 9606:'hsapiens_gene_ensembl'}
    enr_results = defaultdict(lambda: defaultdict(list))

    warnings.simplefilter("ignore")
    
    if gene_sets is None:
        gene_sets = geneset_map[taxid]

    if isinstance(X, list):
        pass
    _blockPrint()
    for cluster, nets in X.items():
        if len(nets)==0:
            continue
        if len(nets[0])==0:
            continue
        for m,net in enumerate(nets[0]):
            if type == "DDI":
                genelist = [mapping[x.split("/")[0]] for x in list(net.nodes) if x.split("/")[0] in set(mapping.keys())]
            else:
                genelist = [mapping[x] for x in list(net.nodes) if x in set(mapping.keys())]
                
            
            enr = gp.enrichr(gene_list = genelist,
                            gene_sets = gene_sets,
                            organism=org_map[taxid],
                            cutoff=cutoff,
                            background=bg[taxid]
                            )
            

            try:
                enr_results[cluster][m].append(enr.results[enr.results['Adjusted P-value']<cutoff])
            except:
                continue
            
        time.sleep(2)
        _enablePrint()

    print("---------Gene Set Enrichment Result---------\n")
    print(f"Method: {method} Database: {gene_sets}")
    for u,v in enr_results.items():
        for m,vv in v.items():
            print("Cluster {} Module {}".format(u, m)," found enriched in {} terms.".format(vv[0].shape[0]))
    print("-----END-----")

    return enr_results

def list_genesets(organism):
    """
    Return a list of gene sets database

    Parameters
    -----------
    organism:
        human, mouse, etc.
    """
    return gp.get_library_name(database=organism)




import pandas as pd
import numpy as np
import time
import random
import warnings
import math
import itertools as it

from ..DataSet import DataSet
from ..BioNetwork import BioNetwork
from .._clustering.clusterobj import clusterObj
from .._clustering import similarities as sim

def dunnsindex(clusterObj):
    clusters = clusterObj.index_clusters
    alltimeseries = clusterObj.DataSet.timeserieslist
    simfunc = clusterObj.metric


    #get the min clusters pair similarity (min dist/max corr)
    minmin = sim.clusters_pair_similarity(clusters, simfunc)
    inter_min = np.max(minmin)
    print(np.array(minmin))

    #max intra cluster sim
    #(max dist/min cor)
    intra_max = np.min(sim.intra_clusters_similarity(alltimeseries, clusters, clusterObj._prototype, simfunc))
    print(np.array(intra_max))
    #min/max    
    return inter_min/intra_max


from numpy.distutils.core import setup
import os
import sys

# def configuration(parent_package="", top_path=None):
#     from numpy.distutils.misc_util import Configuration
#     import numpy

#     libraries = []
#     if os.name == "posix":
#         libraries.append("m")

#     config = Configuration("spycone", parent_package, top_path)

#     config.add_subpackage("_clustering")
#     config.add_subpackage("_connectivity")
#     config.add_subpackage("_clustering")
#     config.add_subpackage("_clustering")

#     return config

# if __name__ == "__main__":
#     setup(**configuration(top_path="").todict())


import numpy as np 
import gc


class pvalue_correction():
    def __init__(self, pvals, method, alpha = 0.05):
        self.pvals= np.sort(pvals) # sort pvals in ascending order
        self.method = method
        self.alpha = alpha
        self.n = len(pvals)
        self.run()
        

    def run(self):
        METHODS = {'bonf': self.bonferroni, 'holm_bonf': self.holm_bonferroni, 'fdr_bh':self.fdr_bh}
        func = METHODS[self.method]
        func()
    
    def bonferroni(self):
        self.corrected_pvals  = self.pvals * self.n
        self.corrected_pvals[self.corrected_pvals>1]=1

    def holm_bonferroni(self):
        notreject = self.pvals >= (self.alpha / np.arange(self.n, 0, -1))
        pre_pvals_corrected = self.pvals * np.arange(self.n, 0, -1)
        self.corrected_pvals  = np.maximum.accumulate(pre_pvals_corrected)
        self.corrected_pvals[self.corrected_pvals>1]=1
        del pre_pvals_corrected

    def fdr_bh(self):
        conf = np.arange(self.n, 0,-1)/self.n
        pre_pvals_corrected = self.pvals * conf
        self.corrected_pvals  = np.maximum.accumulate(pre_pvals_corrected)
        self.corrected_pvals[self.corrected_pvals>1]=1



import pandas as pd
import numpy as np
import time
import random
import warnings
import itertools as it

from ..DataSet import DataSet
from ..BioNetwork import BioNetwork

class shuffling():
    def __init__(self, timeserieslist=None, shuffle_timeseries=None):
        self.timeserieslist = timeserieslist
        # self.shuffletype = shuffletype
        self.shuffle_timeseries = shuffle_timeseries
  
    def shuffle_dataset_rowwise(self):
        '''
        shuffle time series within one time series 
        '''
        alltimeseriesobj = self.timeserieslist

        for X in alltimeseriesobj:
            for Y in X:
                np.random.shuffle(Y)

        return alltimeseriesobj

    def shuffle_globally(self):
        '''
        shuffle time series within the dataset
        '''
  
        alltimeseriesobj = self.timeserieslist
        objectsize = alltimeseriesobj.shape[1]
        timepoints = alltimeseriesobj.shape[2]
        replicates = alltimeseriesobj.shape[0]

        eachrow = np.zeros((objectsize, timepoints), dtype="double")
        shuffled_rows = np.zeros((objectsize, timepoints), dtype="double")
        shuffled = np.zeros((replicates, objectsize, timepoints), dtype="double")
        randomids = np.random.randint(objectsize*timepoints, size=objectsize*timepoints)

        #cython
 
        for r in range(replicates):
            eachrow = np.zeros((objectsize, timepoints), dtype="double")
            for o in range(objectsize):
                for tp in range(timepoints):
                    randomid = random.choice(randomids)
                    randomobj = randomid // timepoints
                    randomtp = randomid % timepoints
                    eachrow[o][tp] = alltimeseriesobj[r,randomobj,randomtp]

                shuffled_rows[o] = eachrow[o]


            shuffled[r] = shuffled_rows
     
        #multiprocessing here
        # eachrow = [alltimeseriesobj[:,i,:] for i in range(alltimeseriesobj.shape[1])]
       
        # shuffled_rows = []
        # with ProcessPoolExecutor() as exe:
        #     for row in exe.map(self._permute_row, eachrow):
        #         shuffled_rows.append(row)
        
        # #swap axes of the np.array for 0 to 1
        # shuffled_rows = np.swapaxes(shuffled_rows, 0, 1)

        # shuffled = self._shufflearray(shuffled_rows)

        self.shuffle_timeseries = shuffled 

        return shuffled



from .BioNetwork import BioNetwork
from .DataSet import DataSet as dataset
from .clustering import clustering
from .iso import iso_function
from .preprocess import preprocess
from .visualize import vis_all_clusters,  switch_plot, gsea_plot, vis_modules, vis_better_modules, compare_mod
from .go_terms import clusters_gsea, modules_gsea, list_genesets, list_gsea
#from .connectivity import connectivity
from .run_domino import run_domino, run_domain_domino
from .DOMINO.src.core import domino
from .splicingfactor import SF_coexpression, SF_motifsearch
#from ._NEASE import nease


from networkx.algorithms.cluster import clustering
import pandas as pd
import numpy as np
import time
import random
import warnings
import networkx as nx
import matplotlib.pyplot as plt
import matplotlib
from matplotlib import pylab as pl
import seaborn as sns
from collections import defaultdict


from .DataSet import DataSet
from .BioNetwork import BioNetwork
from ._clustering.clusterobj import clusterObj
from ._connectivity.connectivity_task import get_enrichment_undirected_edges_between_clusters_nodedegrees, get_total_undirected_edge_prob_nodedegrees, get_enrichment_undirected_edges_between_clustering_nodedegrees

class connectivity():
    def __init__(self,
                clusterObj,
                DataSet,
                BioNetwork,
                clusterObj2 = None):
        self.clusterObj = clusterObj
        self.DataSet = DataSet
        self.BioNetwork = BioNetwork
        self.directed_enrichment = None
        self.undirected_enrichment = None
        self.total_undirected_edge_prob_nodedegrees = None
        self.clusterObj2 = clusterObj2


    def calculate_connectivity(self):
        clusteringresults = self.clusterObj.genelist_clusters
        dataset = self.DataSet

        ##precalculate network attributes
        maxdegree = self.BioNetwork._get_maxdegree()
        edgecounts = self.BioNetwork._get_total_undirected_edge_count_nodedegrees()
        connected_nodes = self.BioNetwork._get_connected_nodearray()
        nodedegrees = self.BioNetwork._get_undirected_node_degree()
        total_nodedeg_count = self.BioNetwork._get_total_node_degree_counts()

        if self.clusterObj2 is None:
            print("Calculating edges probabilities between clusters.")
            if self.total_undirected_edge_prob_nodedegrees is None:
                c = get_total_undirected_edge_prob_nodedegrees(edgecounts, maxdegree, total_nodedeg_count)
                self.total_undirected_edge_prob_nodedegrees = c
            else:
                c = self.total_undirected_edge_prob_nodedegrees

            print("Calculating enriched edges between clusters.")
            d = get_enrichment_undirected_edges_between_clusters_nodedegrees(c, clusteringresults, maxdegree, edgecounts, connected_nodes, nodedegrees)
            
            ###normalize the enrichment odd ratio with number of object in both clusters
            newv = defaultdict(float)
            for u,v in d.items():
                cluster1 = int(u.split(",")[0])
                cluster2 = int(u.split(",")[1])
                totaln = len(clusteringresults[cluster1])+len(clusteringresults[cluster2])
                newv[u] = v/totaln

            self.undirected_enrichment = newv
            print("Calculated enrichment of edges between clusters")
        
        else:
            clusteringresults2 = self.clusterObj2.genelist_clusters
            print("Calculating edges probabilities between clusters.")
            if self.total_undirected_edge_prob_nodedegrees is None:
                c = get_total_undirected_edge_prob_nodedegrees(edgecounts, maxdegree, total_nodedeg_count)
                self.total_undirected_edge_prob_nodedegrees = c
            else:
                c = self.total_undirected_edge_prob_nodedegrees

            print("Calculating enriched edges between clusters.")
            d = get_enrichment_undirected_edges_between_clustering_nodedegrees(c, clusteringresults, clusteringresults2, maxdegree, edgecounts, connected_nodes, nodedegrees)
            
            ###normalize the enrichment odd ratio with number of object in both clusters
            # newv = defaultdict(float)
            # for u,v in d.items():
            #     cluster1 = int(u.split(",")[0])
            #     cluster2 = int(u.split(",")[1])
            #     totaln = len(clusteringresults[cluster1])+len(clusteringresults2[cluster2])
            #     newv[u] = v/totaln

            self.undirected_enrichment = d
            print("Calculated enrichment of edges between clusters")

        return d

    def vis_cluster_network(self):
        logodd = self.undirected_enrichment
        prototypes = self.clusterObj._prototype

        #create edge list
        logodd_mat = np.zeros((len(prototypes), len(prototypes)))
        prototype_corr = np.zeros((len(prototypes), len(prototypes)))
        for cs, lodd in logodd.items():
            css = cs.split(",")
            logodd_mat[int(css[0])-1,int(css[1])-1] = lodd
            proto1 = prototypes[int(css[0])]
            proto2 = prototypes[int(css[1])]
            prototype_corr[int(css[0])-1,int(css[1])-1] = np.corrcoef(proto1, proto2)[0,1]
            
        x= ['Cluster {}'.format(x) for x in list(self.clusterObj.genelist_clusters.keys())]
        y= ['Cluster {}'.format(x) for x in list(self.clusterObj.genelist_clusters.keys())]

        #code from matplotlib to create heatmap
        #
        # def heatmap(data, row_labels, col_labels, ax=None,
        #         cbar_kw={}, cbarlabel="", **kwargs):
        #     """
        #     Create a heatmap from a numpy array and two lists of labels.

        #     Parameters
        #     ----------
        #     data
        #         A 2D numpy array of shape (N, M).
        #     row_labels
        #         A list or array of length N with the labels for the rows.
        #     col_labels
        #         A list or array of length M with the labels for the columns.
        #     ax
        #         A `matplotlib.axes.Axes` instance to which the heatmap is plotted.  If
        #         not provided, use current axes or create a new one.  Optional.
        #     cbar_kw
        #         A dictionary with arguments to `matplotlib.Figure.colorbar`.  Optional.
        #     cbarlabel
        #         The label for the colorbar.  Optional.
        #     **kwargs
        #         All other arguments are forwarded to `imshow`.
        #     """

        #     if not ax:
        #         ax = plt.gca()

        #     # Plot the heatmap
        #     im = ax.imshow(data, **kwargs)

        #     # Create colorbar
        #     cbar = ax.figure.colorbar(im, ax=ax, **cbar_kw)
        #     cbar.ax.set_ylabel(cbarlabel, rotation=-90, va="bottom")

        #     # We want to show all ticks...
        #     ax.set_xticks(np.arange(data.shape[1]))
        #     ax.set_yticks(np.arange(data.shape[0]))
        #     # ... and label them with the respective list entries.
        #     ax.set_xticklabels(col_labels)
        #     ax.set_yticklabels(row_labels)

        #     # Let the horizontal axes labeling appear on top.
        #     ax.tick_params(top=True, bottom=False,
        #                 labeltop=True, labelbottom=False)

        #     # Rotate the tick labels and set their alignment.
        #     plt.setp(ax.get_xticklabels(), rotation=-30, ha="right",
        #             rotation_mode="anchor")

        #     # Turn spines off and create white grid.
        #     for edge, spine in ax.spines.items():
        #         spine.set_visible(False)

        #     ax.set_xticks(np.arange(data.shape[1]+1)-.5, minor=True)
        #     ax.set_yticks(np.arange(data.shape[0]+1)-.5, minor=True)
        #     ax.grid(which="minor", color="w", linestyle='-', linewidth=3)
        #     ax.tick_params(which="minor", bottom=False, left=False)

        #     return im, cbar

        # def annotate_heatmap(im, data=None, valfmt="{x:.2f}",
        #              textcolors=("black", "white"),
        #              threshold=None, **textkw):
        #     """
        #     A function to annotate a heatmap.

        #     Parameters
        #     ----------
        #     im
        #         The AxesImage to be labeled.
        #     data
        #         Data used to annotate.  If None, the image's data is used.  Optional.
        #     valfmt
        #         The format of the annotations inside the heatmap.  This should either
        #         use the string format method, e.g. "$ {x:.2f}", or be a
        #         `matplotlib.ticker.Formatter`.  Optional.
        #     textcolors
        #         A pair of colors.  The first is used for values below a threshold,
        #         the second for those above.  Optional.
        #     threshold
        #         Value in data units according to which the colors from textcolors are
        #         applied.  If None (the default) uses the middle of the colormap as
        #         separation.  Optional.
        #     **kwargs
        #         All other arguments are forwarded to each call to `text` used to create
        #         the text labels.
        #     """

        #     if not isinstance(data, (list, np.ndarray)):
        #         data = im.get_array()

        #     # Normalize the threshold to the images color range.
        #     if threshold is not None:
        #         threshold = im.norm(threshold)
        #     else:
        #         threshold = im.norm(data.max())/2.

        #     # Set default alignment to center, but allow it to be
        #     # overwritten by textkw.
        #     kw = dict(horizontalalignment="center",
        #             verticalalignment="center")
        #     kw.update(textkw)

        #     # Get the formatter in case a string is supplied
        #     if isinstance(valfmt, str):
        #         valfmt = matplotlib.ticker.StrMethodFormatter(valfmt)

        #     # Loop over the data and create a `Text` for each "pixel".
        #     # Change the text's color depending on the data.
        #     texts = []
        #     for i in range(data.shape[0]):
        #         for j in range(data.shape[1]):
        #             kw.update(color=textcolors[int(im.norm(data[i, j]) > threshold)])
        #             text = im.axes.text(j, i, valfmt(data[i, j], None), **kw)
        #             texts.append(text)

        #     return texts

        # fig, (ax,ax1) = plt.subplots(2,1, figsize=(10,12))

        # im, _ = heatmap(logodd_mat, y, x, ax=ax, cmap="PiYG")
        # texts = annotate_heatmap(im)

        # im, _ = heatmap(prototype_corr, y, x, ax=ax1, cmap="PiYG")
        # texts = annotate_heatmap(im)

        # fig.tight_layout()
        # plt.show()
        colmap = sns.light_palette("seagreen", as_cmap=True)
        sns.clustermap(logodd_mat, xticklabels=x, yticklabels=y, cmap=colmap)
        sns.clustermap(prototype_corr, xticklabels=x, yticklabels=y, cmap='coolwarm')

        return None

    
    





from setuptools import setup, find_packages, dist
import os
from distutils.core import Extension
from distutils.command.sdist import sdist as _sdist
from setuptools import Extension
import sys


def setup_package():
    metadata = dict(name='spycone',
    version='0.0.3',
    description='A splicing-aware time course network enricher',
    url='https://github.com/yollct/spycone.git',
    author='Chit Tong Lio',
    author_email='ct7298@gmail.com',
    license='GPLv3',
    classifiers=[
      "Programming Language :: Python :: 3",
      "Programming Language :: Python :: 3.7",
      "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
      "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    packages=find_packages(),
    include_package_data=True,
    python_requires='>=3.7',
    install_requires=[
      'pandas>=1.0.1',
      'numpy>=1.18.1',
      'seaborn',
      'scikit-learn>=0.23.2',
      'scikit-learn-extra>=0.1.0',
      'networkx>=2.4',
      'matplotlib>=3.1.3',
      'gseapy>=0.10.4',
      'scipy>=1.4.1',
      'statsmodels>=0.11.0',
      'tslearn>=0.5.1.0',
      'python-louvain',
      'plotly>=4.14.3',
      'biopython',
      'gtfparse',  
      'joblib',
      'nease'])
  
    cmdclass = {}
    ext_modules = []
    metadata['setup_requires']=['numpy']

    setup(**metadata)


if __name__ == "__main__":
  setup_package()




import sys
sys.path.insert(0, '../../')
import numpy as np
import os
import time
import shutil
import json
import pandas as pd

from .. import constants
from ..utils.scripts import format_script
from ..utils.ensembl2gene_symbol import  e2g_convertor
import zipfile

import multiprocessing
from functools import reduce
SH_MODULE_NAME = "module"
SH_NUM_GENES = "#_genes"
SH_ENRICHED = "enriched_groups"
SH_DETAILS = "more_details"

SH_TABLE_HEADERS = [SH_MODULE_NAME, SH_NUM_GENES, SH_ENRICHED, SH_DETAILS]

MODULE_TH = 10

def zipdir(path_to_zip, zip_file_path):
    ziph = zipfile.ZipFile(zip_file_path, 'w', zipfile.ZIP_DEFLATED)
    for root, dirs, files in os.walk(path_to_zip):
        for file in files:
            ziph.write(os.path.join(root, file))

def get_network_genes(network_file_name):
    network_df = pd.read_csv(network_file_name, sep="\t")
    src = np.array(network_df.loc[:, 0])
    dst = np.array(network_df.loc[:, 2])
    vertices = list(set(np.append(src, dst)))
    return vertices

def remove_subgraph_self_loops(nodes_to_remove, network_file_name):
    if len(nodes_to_remove) == 0:
        return network_file_name
    network_df = pd.read_csv(network_file_name, sep="\t")
    filtered_network = network_df[network_df.loc[:,0]!=network_df.loc[:,2]]
    new_file_name = os.path.splitext(network_file_name) + "_no_loops" +".sif"
    filtered_network.to_csv(new_file_name, sep="\t", index=False)
    return filtered_network

def remove_subgraph_by_nodes(nodes_to_remove, network_file_name, ts=str(time.time())):
    if len(nodes_to_remove) == 0:
        return network_file_name
    network_df = pd.read_csv(network_file_name, sep="\t")
    filtered_network = network_df[~(network_df.loc[:,0].isin(nodes_to_remove) | network_df.loc[:,2].isin(nodes_to_remove))]
    new_file_name = os.path.splitext(network_file_name)[0] + ts +".sif"
    filtered_network.to_csv(new_file_name, sep="\t", index=False)
    return new_file_name


def summary_intergrative_reports(all_hg_reports, modules_summary, total_hg_report, algo_name, module_genes, report_file_name, dataset_name):
    general_algo_report(algo_name, all_hg_reports, module_genes, modules_summary, report_file_name, total_hg_report, dataset_name)


def disease_algo_report(algo_name, disease_name, expected_genes, module_genes, modules_summary, report_file_name, dataset_name):

    disease_data = {
        "disease_name": disease_name,
        "num_of_modules": len(modules_summary),
        "TP+FN_(_true_)": len(expected_genes),
        "TP+TN_(_retrieved_)": len(module_genes),
        "TP/(TP+TN)_(_precision_)": 0,
        "TP/(TP+FN)_(_recall_)": 0,
        "F1": 0,
        "TP": 0,
        "module_size_avg" : 0,
        "module_size_std" :0
    }
    if len(modules_summary) > 0:
        modules_summary = pd.DataFrame(modules_summary)
        disease_genes_extracted = float(len(set(module_genes).intersection(expected_genes)))
        disease_data["TP"] = disease_genes_extracted
        disease_data["TP/(TP+TN)_(_precision_)"] = disease_genes_extracted / len(module_genes)
        disease_data["TP/(TP+FN)_(_recall_)"] = disease_genes_extracted / len(expected_genes)
        if (disease_data["TP/(TP+TN)_(_precision_)"] + disease_data["TP/(TP+FN)_(_recall_)"]) == 0:
            disease_data["F1"] = 0
        else:
            disease_data["F1"] = 2 * ((disease_data["TP/(TP+TN)_(_precision_)"] * disease_data["TP/(TP+FN)_(_recall_)"]) /
                                  (disease_data["TP/(TP+TN)_(_precision_)"] + disease_data["TP/(TP+FN)_(_recall_)"]))

        disease_data["module_size_avg"] = modules_summary[SH_NUM_GENES].mean()
        disease_data["module_size_std"] = modules_summary[SH_NUM_GENES].std()


    pd.DataFrame([disease_data]).to_csv(os.path.join(constants.OUTPUT_GLOBAL_DIR, dataset_name, algo_name, "{}_disease.tsv".format(report_file_name)),sep="\t", index=False)


def general_algo_report(algo_name, all_hg_reports, module_genes, modules_summary, report_file_name, total_hg_report, dataset_name):
    data = {}
    if len(modules_summary) > 0 :
        df_summary = pd.DataFrame(modules_summary)
        data = {"num_of_modules": df_summary.index.size,
                "module_size_avg": df_summary[SH_NUM_GENES].mean(),
                "module_size_std": df_summary[SH_NUM_GENES].std(),
                "total_num_genes": len(module_genes)
                }

    df = pd.DataFrame()
    if len(data) >0:
        df = pd.DataFrame([data])

    df.to_csv(
        os.path.join(constants.OUTPUT_GLOBAL_DIR, dataset_name, algo_name,
                     "{}_general.tsv".format(report_file_name)), sep="\t", index=False)



# def output_modules(output_file_name, modules, score_file_name, output_base_dir=""):
#     output_data = create_modules_output(modules, score_file_name)
#     file(output_file_name, 'w+').write(output_base_dir + "\n")
#     json.dump(output_data, file(output_file_name, 'a+'))
#     sys.stdout.write(output_file_name)

def reduce_to_dict(x,y):
    if y["id"] in x:
        x[y["id"]]["modules"] = x[y["id"]]["modules"] + y["modules"]
    else:
        x[y["id"]]=y
    return x

def merge_two_dicts(x, y):

    z = x.copy()
    z.update(y)
    return z

def create_modules_output(G_modules, score_file_name):
    scores=None
    if score_file_name is not None:
       print("score_file_name: {}".format(score_file_name))
       print(pd.read_csv(score_file_name,sep="\t").columns)
       scores = pd.read_csv(score_file_name,sep="\t").set_index("id")

       if constants.IS_PVAL_SCORES:
            scores["score"] = scores["pval"].apply(lambda x: -np.log10(x))

    zero_scores = [ {"score" : 0, "id" : gene} for G_module in G_modules for gene in G_module.nodes if scores is None or gene not in scores.index]
    if len(zero_scores) !=0:
        zero_scores = pd.DataFrame(zero_scores).set_index("id")
        zero_scores=zero_scores[~zero_scores.index.duplicated(keep='first')]
        scores = pd.concat([scores, zero_scores],axis=0)
    return [merge_two_dicts({"id" : k}, v) for k,v in reduce(reduce_to_dict, [{"eid": gene, "modules": [i], "id": gene, "gene_symbol": e2g_convertor([gene])[0], "score" : float(scores.loc[gene,"score"])} for i, G_module in enumerate(G_modules) for gene in G_module],\
            {}).items()]

def draw_network(G_modules, score_file_name, network_file_name):
    output = [{"data" : x, "label" : x["eid"], "selected" : True } for x in create_modules_output(G_modules, score_file_name)]
    # active_genes = [y for x in G_modules for y in x.nodes]
    active_edges = [y for x in G_modules for y in x.edges]
    # active_edges = [[x.iloc[0], x.iloc[2]] for i, x in pd.read_csv(network_file_name, sep="\t").iterrows() if x.iloc[0] in active_genes and x.iloc[2] in active_genes]
    additional_edges = [] # [[x.iloc[0], x.iloc[2]] for i, x in pd.read_csv(network_file_name, sep="\t").iterrows() if not (x.iloc[0] in active_genes and x.iloc[2] in active_genes) and (x.iloc[0] in active_genes or x.iloc[2] in active_genes)]
    # additional_nodes = [] # [y for x in (active_edges + additional_edges) for y in x if y if y not in active_genes]
    additional_nodes = [] # list(set(additional_nodes))

    return output + [{"data" : {"id" : x, "eid" : x, "modules" : []}, "label" : ""} for x in additional_nodes] + [{"data": {"id" : x[0]+"_"+x[1], "source":x[0], "target":x[1]}, "label" : ""} for x in additional_edges] + [{"data": {"id" : x[0]+"_"+x[1], "source":x[0], "target":x[1]}, "label" : "-"} for x in active_edges]



def generate_report_from_template(cy, output_base_dir, output_file_name):

    len([x for x in cy if not "source" in x["data"] and len(x["data"]["modules"])>0])
    report_file_name=format_script(os.path.join(os.path.dirname(os.path.abspath(__file__)),'../data', "graph.html"), NUM_OF_GENES=len([x for x in cy if not "source" in x["data"] and len(x["data"]["modules"])>0]), HG_REPORT=[], MODULES_SUMMARY=[], DISEASE_GENES=[], DATA=json.dumps(cy))

    shutil.move(report_file_name,
                os.path.join(output_base_dir, "module_{}.html".format(output_file_name)))
    return "module_{}.html".format(output_file_name)


def visualize_modules(dataset_name, G_modules, score_file_name, network_file_name, output_base_dir):
    print("visualizing modules...")
    if not os.path.exists(output_base_dir):
        os.makedirs(output_base_dir)

    manager=multiprocessing.Manager()
    modules_summary = manager.list()

    params=[]
    for i, G_module in enumerate(G_modules):
        params.append([i, G_module, score_file_name, network_file_name, dataset_name, modules_summary, output_base_dir])
    p=multiprocessing.Pool(constants.N_OF_THREADS)
    p.map(module_report, params)
    p.close()
    # [module_report(p) for p in params]

def module_report(params):
    module_index, G_module, score_file_name, network_file_name, dataset_name, modules_summary, output_base_dir=params
    print("visualize module {} for dataset {}".format(module_index, dataset_name))

    modules_summary_row = {SH_MODULE_NAME: module_index, SH_NUM_GENES: len(G_module.nodes)}
    cy = draw_network([G_module], score_file_name, network_file_name)

    generate_report_from_template(cy, output_base_dir, str(module_index))
    if modules_summary is not None:
        modules_summary.append(modules_summary_row)
    return modules_summary





import pandas as pd
import numpy as np
import time
import random
import warnings
import sys
from scipy.stats import chi2

sys.path.insert(0, "../")
from ..DataSet import DataSet
from ..BioNetwork import BioNetwork
from .._feature import get_features as fea
from .._fitness.fitness_score import fitness_score
from .._shuffle import shuffle

from .. import clustering

class basic_pvalue():
    def __init__(self, testobject, object_type, iso_switch = True, shuffle=None, fitness_scores = None, fitness_scores_two_sided = True, n_permutations=1000):
        self.testobject=testobject ##DataSet
        self.object_type = object_type #connectivity, clusters, iso_pairs
        self.shuffle = shuffle
        self.fitness_scores = fitness_scores
        self.fitness_scores_two_sided = fitness_scores_two_sided
        #self.combine_pvalues = combine_pvalues
        self.n_permutations = n_permutations
        self.iso_switch = iso_switch

    def initialize_features_for_permutation(self, features, shuffled):
        pass

    def calculate_original_object_fitness(self):
        pass

    # Print iterations progress
    def progressbar(self, it, prefix="", size=60, file=sys.stdout):
        count = len(it)
        def show(j):
            x = int(size*j/count)
            file.write("%s[%s%s] %i/%i\r" % (prefix, "#"*x, "."*int((size-x)), j, count))
            file.flush()        
        show(0)
        for i, item in enumerate(it):
            yield item
            show(i+1)
        file.write("\n")
        file.flush()


    def empirical_pvalue(self, total_observation,better_observation):
        if total_observation > 0:
            result = (1+better_observation)/ (1+total_observation)
            return result
        
    def combine_pvals(self, pvals):
        chisqprob = lambda chisq, df: chi2.sf(chisq, df)
        s = -2 * np.sum(np.log(pvals))
        
        return chisqprob(s, 2 * len(pvals))

        
    def do_calculate(self):
        if self.object_type=="clusters":
            n_object = self.testobject.clusterobj._final_n_cluster
            timeserieslist = self.testobject.timeserieslist
            #get features of clusters
            print("Calculating features.")
            features_original = fea.featuresObj(self.testobject, timeserieslist, feature_type="clusters")
            #get fitness of clusters
            print("Calculating fitness score.")
            original_fitness_score = features_original.feature_store
            print(original_fitness_score)

            #data structure for storing permuted features #[features][permutation]
            permuted_fitness = np.zeros((original_fitness_score.shape[0], n_object))

        # else:
        #     if not isinstance(testobject, clusterObj):
        #         raise ValueError("ClusterObj object is needed.")
            
        if self.object_type == "iso_pairs":
            n_object = len(self.testobject.isoobj.isopairs['major_transcript'])
            timeserieslist = self.testobject.timeserieslist
            features_original = fea.featuresObj(self.testobject, timeserieslist, feature_type="iso_pairs")
            original_fitness_score = features_original.feature_store
            permuted_fitness = np.zeros((original_fitness_score.shape[0], n_object))
 

        batch_smaller=np.zeros((original_fitness_score.shape[0], n_object))
        batch_larger=np.zeros((original_fitness_score.shape[0], n_object))
        batch_equal=np.zeros((original_fitness_score.shape[0], n_object))
        batch_comp=np.zeros((original_fitness_score.shape[0], n_object))

        #shuffle feature and fitness  

        #def _cal_permutations(p):
        for p in self.progressbar(it=range(self.n_permutations), prefix="Permutation:", size=self.n_permutations/10):
            shuffleObj = shuffle.shuffling(timeserieslist)
            shuffle1 = np.array(shuffleObj.shuffle_dataset_rowwise(), dtype="double")

            self.testobject.timeserieslist = shuffle1 ##reassign testobject time series to the shuffled dataset
            #calculate features for shuffle obj
            iso = iso_function(self.testobject)
            corrdf = iso.detect_isoform_switch(filtering=False, min_diff=self.testobject.isoobj.min_diff, corr_cutoff=self.testobject.isoobj.corr_cutoff, event_im_cutoff=self.testobject.isoobj.event_im_cutoff)
            self.testobject = iso.total_isoform_usage(corrdf)
            shuffle_clu = clustering.clustering(self.testobject, input_type="isoformusage", algorithm=self.testobject.clusterobj.algorithm, metric=self.testobject.clusterobj.metric, linkage=self.testobject.clusterobj.linkage, n_clusters=self.testobject.clusterobj.n_clusters)


            if self.object_type=="clusters":
                permuted_features = fea.featuresObj(testobject=self.testobject, timeserieslist=shuffle1, feature_type = "clusters")
            #calculate fitness for shuffle obj 
                permuted_fitness = permuted_features.feature_store


            elif self.object_type=="iso_pairs":
                permuted_features = fea.featuresObj(testobject=self.testobject, timeserieslist = shuffle1, feature_type="shuffle_iso_pairs")
                permuted_fitness = permuted_features.feature_store

            
            for fs in range(original_fitness_score.shape[0]):
                fitness = original_fitness_score[fs]
                shuffledfitness = permuted_fitness[fs]
                    
                #comparison
                for o in range(len(fitness)):
                    compare = shuffledfitness[o] - fitness[o]

                    if compare < 0:
                        batch_smaller[fs][o] += 1
                    elif compare > 0:
                        batch_larger[fs][o] += 1
                    else:
                        batch_equal[fs][o] += 1
                    batch_comp[fs][o] += 1

            p+=1
            time.sleep(0.1)

        #res = Parallel(backend="multiprocessing") (delayed(_cal_permutations)(int(p)) for p in self.progressbar(it=range(self.n_permutations), prefix="Permutation:", size=self.n_permutations/10))
        #print(res)
       
        #end permutation
        print("Done permutation")
        pvals=np.zeros((original_fitness_score.shape[0], n_object))
        total_observation = 0
        better_observation = 0
        

        for fs in range(original_fitness_score.shape[0]):
            for o in range(n_object):
                total_observation = batch_comp[fs][o]
                

                if self.fitness_scores_two_sided is True:
                    better_observation = min(batch_larger[fs][o], batch_smaller[fs][o])*2 + batch_equal[fs][o]
                else:
                    better_observation = batch_equal[fs][o] + batch_larger[fs][o]

                pvals[fs][o] = self.empirical_pvalue(total_observation, better_observation)
                
        # p_vals = Parallel(backend="multiprocessing") (delayed(_count_permutations)(fs, o) for o in range(n_object) for fs in range(original_fitness_score.shape[0])) 
       
        #p_vals = np.array(p_vals).reshape(n_object, original_fitness_score.shape[0])

        ##combine pvalues
        result_pvalue = np.zeros(n_object)

        for o in range(n_object):
            p_values_list = np.zeros(original_fitness_score.shape[0])
            p_values_list = pvals[:,o]
            result_pvalue[o] = self.combine_pvals(p_values_list)

        return result_pvalue

        
        




import sys
sys.path.insert(0, "../")

import random
import os

import pandas as pd
import numpy as np
import pickle
import multiprocessing
import matplotlib
matplotlib.use("Agg")
from scipy.stats import hypergeom
from statsmodels.sandbox.stats.multicomp import fdrcorrection0

import networkx as nx
from networkx.algorithms.community.quality import modularity
from networkx.algorithms.community.centrality import girvan_newman
from networkx.algorithms.components import connected_components

from functools import reduce
from ..utils.graph_influence_linear_th import linear_threshold
from .preprocess_slices import read_preprocessed_slices
from .network_builder import build_network
from ..constants import * 

import pcst_fast
G_modularity = None



def extract_scores(gene_list, pval=None):
    """"""
    # scores1 = pd.read_csv("./ticone_pkg/ticone/DOMINO/examples/tnfa_active_genes_file.txt", sep='\t', index_col=0, header=None, dtype=str)

    ##
    scores = pd.DataFrame(index=gene_list, dtype=str)
    ## 
    if pval:
        scores["score"] = pval
    else:
        scores["score"] = 1

    return scores


def add_scores_to_nodes(G, scores):
    """"""
    inds = []
    for nd in G.nodes:
        G.nodes[nd]["pertubed_node"] = False
        G.nodes[nd]["score"] = 0

    for ind, row in scores.iterrows():
        if ind in G.nodes:
            inds.append(ind)
            G.nodes[ind]["score"] = row["score"]
            
            G.nodes[ind]["pertubed_node"] = row["score"] > 0  # binarizing the activeness

    #print("check network int", G.nodes[ind]['score'])
    return G


def create_subgraph(params):
    cur_module = params
    global G_modularity
    nodes = set(cur_module)
    res = G_modularity.subgraph(list(nodes))
    return res


def prune_network_by_modularity(G, modules):
    global G_modularity
    # if os.path.exists(cache_file) and USE_CACHE:
    #     #print(f'fetch cache file for subnetworks {cache_file}')
    #     G_modularity = pickle.load(open(cache_file, 'rb'))
    #     for n in G_modularity:
    #         G_modularity.nodes[n]['pertubed_node'] = G.nodes[n]['pertubed_node']
    #     #print('pkl is loaded')
    #     return G_modularity

    #print(f'generating subgraphs...')
    G_modularity = G
    #print(f"Before slicing: n of cc:{len(list(connected_components(G_modularity)))}, n of nodes: {len(G_modularity.nodes)}, n of edges, {len(G_modularity.edges)#}")
    p = multiprocessing.Pool(N_OF_THREADS)

    G_modules = p.map(create_subgraph, [m for m in modules])
    p.close()
    # print(f'{modules}')
    #print(f'# of modules after extraction: {len(G_modules)}')
    G_modularity = nx.algorithms.operators.union_all(G_modules)
    #print(
    #    f"After slicing: n of cc:{len(list(connected_components(G_modularity)))}, n of nodes: {len(G_modularity.nodes)}, n of edges, {len(G_modularity.edges)}")
    #pickle.dump(G_modularity, open(cache_file, 'wb+'))
    #print('subgraphs\' pkl is saved')


def prune_network_by_modularity_old(G, modules, dummy):
    G_modularity = G.copy()
    edges_to_remove = []
    for cur_edge in G_modularity.copy().edges:
        in_cc = False
        for cur_module in modules:
            if cur_edge[0] in cur_module and cur_edge[1] in cur_module:
                in_cc = True
        if not in_cc:
            edges_to_remove.append(cur_edge)

    G_modularity.remove_edges_from(edges_to_remove)
    return G_modularity


def get_pcst_prize(G_cc, prize_factor, n_steps):
    prizes = {}
    p_cc = linear_threshold(G_cc, [n for n in G_cc.nodes if G_cc.nodes[n]['pertubed_node'] > 0], steps=n_steps)
    for p_node in G_cc.nodes:
        prizes[p_node] = 0
    for i_cur_layer, cur_layer in enumerate(p_cc):
        for cur_node in cur_layer:
            prizes[cur_node] += prize_factor ** i_cur_layer

    return prizes

def run_pcst(G_cc, i_cc, labels, n_steps, nodes, prize_factor):
    ## set prize ##
    prizes = get_pcst_prize(G_cc, prize_factor, n_steps)
    vertices_prizes = []
    for cur_node in nodes:
        vertices_prizes.append(
            G_cc.nodes[cur_node]["pertubed_node"] if G_cc.nodes[cur_node]["pertubed_node"] else prizes[cur_node])

    ## set cost ##
    edges_grid = []
    for cur_edge in G_cc.edges:
        edges_grid.append([nodes.index(cur_edge[0]), nodes.index(cur_edge[1])])

    edges_costs = []
    for cur_edge in edges_grid:
        u_score = 0 if G_cc.nodes[nodes[cur_edge[0]]]["pertubed_node"] else 0.9999
        v_score = 0 if G_cc.nodes[nodes[cur_edge[1]]]["pertubed_node"] else 0.9999

        edges_costs.append(np.min([u_score, v_score]))

    ## find pcst component by running pcst fast##
    root = -1
    num_clusters = 1
    pruning = 'strong'  # 'none'
    verbosity_level = 0
    vertices, edges = pcst_fast.pcst_fast(edges_grid, vertices_prizes, edges_costs, root, num_clusters, pruning,
                                          verbosity_level)

    return edges, edges_grid


def split_subslice_into_putative_modules(G_optimized, improvement_delta, modularity_score_objective, best_modularity):
    cur_components = [G_optimized.subgraph(c) for c in connected_components(G_optimized)]
    cur_modularity = modularity(G_optimized, cur_components, weight='weight')
    if cur_modularity >= modularity_score_objective:
        return True, best_modularity

        if len(n_nodes) < 4:
            G_optimized.remove_nodes_from(n_nodes)

    cur_components = [G_optimized.subgraph(c) for c in connected_components(G_optimized)]
    if len(cur_components) == 0:
        return True, best_modularity

    optimized_connected_components = girvan_newman(G_optimized)
    cur_components = sorted(next(optimized_connected_components))
    cur_modularity = modularity(G_optimized, cur_components, weight='weight')
    if cur_modularity <= best_modularity + improvement_delta:
        return True, best_modularity

    else:
        optimal_components = cur_components

        edges_to_remove = []
        for cur_edge in G_optimized.edges:
            included = False
            for n_nodes in optimal_components:
                if cur_edge[0] in n_nodes and cur_edge[1] in n_nodes:
                    included = True
            if not included:
                edges_to_remove.append(cur_edge)

        G_optimized.remove_edges_from(edges_to_remove)

        return False, cur_modularity


def get_putative_modules(G, full_G=None, improvement_delta=0, modularity_score_objective=1, module_threshold=0.05,
                         n_cc=1.0):
    """"""

    if full_G == None:
        full_G = G
    G_optimized = G.copy()

    # clean subslice from cycles and isolated nodes
    G_optimized.remove_edges_from(list(nx.selfloop_edges(G_optimized)))
    G_optimized.remove_nodes_from(list(nx.isolates(G_optimized)))

    # check subslice enrichment for active nodes
    pertubed_nodes = [cur_node for cur_node in full_G.nodes if full_G.nodes[cur_node]["pertubed_node"]]
    pertubed_nodes_in_cc = [n for n in G_optimized.nodes if G_optimized.nodes[n]["pertubed_node"]]
    n_nodes = list(G_optimized.nodes)
    sig_score = hypergeom.sf(len(pertubed_nodes_in_cc), len(full_G.nodes), len(pertubed_nodes),
                             len(n_nodes)) \
                + hypergeom.pmf(len(pertubed_nodes_in_cc), len(full_G.nodes), len(pertubed_nodes),
                                len(n_nodes))

    sig_score = sig_score / n_cc

    # if subslice is not enriched for active nodes split in into putative modules. otherwise, report it as a single putative module
    # print(f'{sig_score}<{module_threshold} and {len(G_optimized.nodes)}<30')
    is_enriched_sublice = (len(G_optimized.nodes) < 100) or len(
        G_optimized.nodes) == 0  # sig_score<module_threshold and l

    break_loop = is_enriched_sublice
    best_modularity = -1
    while not break_loop:
        break_loop, best_modularity = split_subslice_into_putative_modules(G_optimized, improvement_delta,
                                                                           modularity_score_objective, best_modularity)

    G_optimized.remove_nodes_from(list(nx.isolates(G_optimized)))

    cc_optimized = [] if len(G_optimized.nodes) == 0 else [G_optimized.subgraph(c) for c in
                                                           connected_components(G_optimized)]

    return G_optimized, cc_optimized


def retain_relevant_slices(G_original, module_sig_th):
    global G_modularity

    pertubed_nodes = []
    for cur_node in G_modularity.nodes():
        if G_modularity.nodes[cur_node]["pertubed_node"]:
            pertubed_nodes.append(cur_node)

    ccs = [G_modularity.subgraph(c) for c in connected_components(G_modularity)]
    params = []
    p = multiprocessing.Pool(N_OF_THREADS)
    n_G_original = len(G_original)
    n_pertubed_nodes = len(pertubed_nodes)
    pertubed_nodes_in_ccs = []
    #print(f"number of slices: {len(list(ccs))}")
    for i_cur_cc, cur_cc in enumerate(ccs):
        pertubed_nodes_in_ccs.append(
            len([cur_node for cur_node in cur_cc if G_modularity.nodes[cur_node]["pertubed_node"]])) 
    perturbation_factor = min(0.7, (float(n_pertubed_nodes) / n_G_original) * (
                1 + 100 / n_G_original ** 0.5))

    for i_cur_cc, cur_cc in enumerate(ccs):
        params.append([n_G_original, cur_cc, i_cur_cc, n_pertubed_nodes, perturbation_factor])

    res = [a for a in p.map(pf_filter, params) if a is not None]
    #print(f'# of slices after perturbation TH: {len(res)}/{len(params)}')
    p.close()
    if len(res) == 0:
        return nx.Graph(), [], []
    large_modules, sig_scores = zip(*res)
    fdr_bh_results = fdrcorrection0(sig_scores, alpha=module_sig_th, method='indep',
                                    is_sorted=False)

    # print(fdr_bh_results)
    # print(f'min: {min(list(fdr_bh_results[1]))}')
    passed_modules = [cur_cc for cur_cc, is_passed_th in zip(large_modules, fdr_bh_results[0]) if is_passed_th]
    return nx.algorithms.operators.union_all(passed_modules) if len(passed_modules) > 0 else nx.Graph(), [list(m.nodes)
                                                                                                          for m in
                                                                                                          passed_modules], \
           fdr_bh_results[1]


def pf_filter(params):
    global G_modularity
    n_G_original, cur_cc, i_cur_cc, n_pertubed_nodes, perturbation_factor = params
    pertubed_nodes_in_cc = [cur_node for cur_node in cur_cc if G_modularity.nodes[cur_node]["pertubed_node"]]
    if len(cur_cc) < 4 or n_pertubed_nodes == 0 or not (
            len(pertubed_nodes_in_cc) / float(len(cur_cc)) >= perturbation_factor or len(pertubed_nodes_in_cc) / float(
            n_pertubed_nodes) >= 0.1):
        return None
    else:
        score = hypergeom.sf(len(pertubed_nodes_in_cc), n_G_original, n_pertubed_nodes,
                             len(cur_cc)) \
                + hypergeom.pmf(len(pertubed_nodes_in_cc), n_G_original, n_pertubed_nodes,
                                len(cur_cc))
        return (cur_cc, score)


def analyze_slice(params):
    G, cc, i_cc, n_steps, relevant_slices, prize_factor, module_threshold = params
    G_cc = nx.subgraph(G, cc)
    nodes = list(G_cc.nodes)
    labels = {n: G_cc.nodes[n] for n in nodes}
    n_pertubed_nodes = sum([G.nodes[a]["pertubed_node"] for a in G.nodes])
    prize_factor = max(0, 1 - 3 * n_pertubed_nodes / float(len(G.nodes)))
    # print(f'active gene ratio: {n_pertubed_nodes}/{len(G_cc.nodes)}')
    # print(f"prize factor: {prize_factor}")
    edges, edges_grid = run_pcst(G_cc, i_cc, labels, n_steps, nodes, prize_factor)
    G_subslice = nx.Graph()
    G_subslice.add_edges_from([(nodes[edges_grid[e][0]], nodes[edges_grid[e][1]]) for e in edges])
    nx.set_node_attributes(G_subslice, {n: labels[n] for n in G_subslice.nodes})
    modularity_score_objective = np.log(len(G_subslice.nodes)) / np.log(len(G.nodes)) if len(
        G_subslice.nodes) > 10 else -1
    subslice_after_ng, putative_modules_of_slice = get_putative_modules(G_subslice, G, improvement_delta=10 ** -2,
                                                                        modularity_score_objective=modularity_score_objective,
                                                                        n_cc=len(relevant_slices),
                                                                        module_threshold=module_threshold)

    return putative_modules_of_slice


def get_final_modules(G, G_putative_modules):
    module_sigs = []
    sig_scores = []
    for i_cur_module, cur_G_module in enumerate(G_putative_modules):
        pertubed_nodes_in_cc = [cur_node for cur_node in cur_G_module.nodes if G.nodes[cur_node]["pertubed_node"]]
        pertubed_nodes = [cur_node for cur_node in G.nodes if G.nodes[cur_node]["pertubed_node"]]

        sig_score = hypergeom.sf(len(pertubed_nodes_in_cc), len(G.nodes), len(pertubed_nodes),
                                 len(cur_G_module.nodes)) \
                    + hypergeom.pmf(len(pertubed_nodes_in_cc), len(G.nodes), len(pertubed_nodes),
                                    len(cur_G_module.nodes))

        final_module_threshold = 0.05 / len(G_putative_modules)
        if sig_score <= final_module_threshold:
            module_sigs.append((cur_G_module, sig_score / len(G_putative_modules)))
            sig_scores.append(sig_score)

    module_sigs = sorted(module_sigs, key=lambda a: a[1])
    return [a[0] for a in module_sigs], sig_scores


def main(active_genes_file, network_file, scores=None, slices_file=None, slice_threshold=0.3, module_threshold=0.05, prize_factor=0,
         n_steps=20):
    # if os.path.exists(f'{network_file}.pkl') and USE_CACHE:
    #     G = pickle.load(open(f'{network_file}.pkl', 'rb'))
    #     #print(f'network\' pkl is loaded: {network_file}.pkl')
    # else:
        #print(f'generating graph from {network_file}')
    G = build_network(network_file)
        #pickle.dump(G, open(f'{network_file}.pkl', 'wb+'))
        #print(f'network\' pkl is saved: {network_file}.pkl')

    #print("done building network")
    # assign activeness to nodes
    scores = extract_scores(active_genes_file, scores)
    G = add_scores_to_nodes(G, scores)

    modularity_connected_components = read_preprocessed_slices(slices_file)

    global G_modularity
    prune_network_by_modularity(G, modularity_connected_components)
    G_modularity, relevant_slices, qvals = retain_relevant_slices(G, slice_threshold)
    #print(f'{len(relevant_slices)} relevant slices were retained with threshold {slice_threshold}')
    params = []
    for i_cc, cc in enumerate(relevant_slices):
        params.append([G, cc, i_cc, n_steps, relevant_slices, prize_factor, module_threshold])
    p = multiprocessing.Pool(N_OF_THREADS)
    putative_modules = reduce(lambda a, b: a + b, p.map(analyze_slice, params), [])
    p.close()
    #print(f'n of putative modules: {len(putative_modules)}')
    final_modules, sig_scores = get_final_modules(G, putative_modules)
    #print(f'n of final modules: {len(final_modules)} (n={[len(list(m)) for m in final_modules]})')
    return final_modules, sig_scores



import pandas as pd
import numpy as np
import time
import random
import warnings

from ..DataSet import DataSet
from ..BioNetwork import BioNetwork
from .scaler import tanhnormalizer_scaler
from .._feature.get_features import featuresObj

class fitness_score():
    def __init__(self, featureObj):
        self.featureObj = featureObj
        

    def calculate_fitness(self):
        object_type = self.featureObj.feature_type
        result = 0
        feature_store = self.featureObj.feature_store
        final_fvalue=[]

        for f in range(len(feature_store)):
            feature = feature_store[f]
            final_fvalue.append(feature)

        final_fvalue = np.array(final_fvalue)

        return final_fvalue




import pandas as pd
import numpy as np
import time
from functools import reduce
from collections import defaultdict
from sklearn.cluster import *
from sklearn.metrics.cluster import adjusted_rand_score
from sklearn.metrics import pairwise_distances, silhouette_score
from tslearn.metrics import cdist_soft_dtw
from tslearn.clustering import TimeSeriesKMeans
from tslearn.clustering import silhouette_score as tssilhouette_score
from sklearn.metrics import davies_bouldin_score

from ._prototype.prototype_builder import _prototype_builder
#from ._util_stat.qc import _qc_silhouette
from ._clustering.build_cluster import cluster_map_to_item
from ._clustering.kmedoids import KMedoids
from ._clustering.clusterobj import clusterObj
# from ._util_stat import compute_pvalues 

class clustering(clusterObj):
    """
    Clustering object

    Parameters
    -----------
    DataSet
        DataSet object

    BioNetwork (needed if composite is True)
        BioNetwork object

    input_type: str
        clustering expression data put "expression", and clustering total isoform usage put "isoformusage"
    
    algorithm : {'kmeans', 'kmedoids', 'dbscan', 'hierarchical', 'optics'}
        clustering algorithms from sklearn

    composite : `boolean`, default=True
        if True, distance metrics is composited with inverse shortest path
    
    metric : {'euclidean', 'correlation'}
        metrics from sklearn

    linkage : only for 'hierarchical' clustering {default='average', 'complete', 'ward'}

    prototypefunction : {default='median', 'mean'}
        aggregation function for cluster prototypes

    searchspace : (default=20)
        range to search for optimal number of clusters
    
    Attributes
    -----------
    _prototype : dictionary of prototypes for each cluster (keys)
    
    _lables : array with length of object
        The cluster label of each object

    genelist_clusters : dictionary of clusters
        Key and values pair of clustering with entrez ID (gene_list ID)

    index_clusters : dictionary of clusters
        Key and values pair of clustering with indices 

    symbs_clusters : dictionary of clusters
        Key and values pair of clustering with gene symbols 

    _final_n_cluster : 
        number of clusters

    _silhouette_index :
        silhouette index of this clustering
    
    Examples 
    ---------
    

    """
    def __init__(self,
                DataSet,
                algorithm,
                input_type,
                n_clusters=10,
                composite = False,
                BioNetwork=None,
                metric = "euclidean",
                prototypefunction = "median",
                linkage="average",  #scikit learn recommended
                searchspace = 20,
                seed = 1234321,
                transform = None,
                **kwargs
                ):
        self.DataSet = DataSet 
        self.algorithm = algorithm
        self.n_clusters= n_clusters
        self.input_type = input_type ###expression or isoformusage
        self.BioNetwork = BioNetwork
        self.composite = composite
        self.linkage = linkage
        self.metric = metric
        self.prototypefunction = prototypefunction
        self.seed = seed
        self.transform = transform


        self.kwargs = kwargs
        self.params_init = {}
        self.params_init.update((key, value) for key, value in self.kwargs.items())

        if self.algorithm == "kmeans":
            assert self.metric == "euclidean" or self.metric == "soft_dtw", "Kmeans can only use euclidean or soft dtw distances as metric."


    CLUSTERING_ALGORITHMS = {'dbscan':DBSCAN, 'optics':OPTICS, 'kmeans': TimeSeriesKMeans, 'kmedoids':KMedoids, 'hierarchical': AgglomerativeClustering}
    PROTOTYPE = {'median':np.median, 'mean':np.mean}
    TRANSFORMATION = {'log':np.log10}
    
    def find_clusters(self):
        """
        Perform clustering 
        """
        t1 = time.time()
        
        if self.input_type == "expression":
            alltimeseriesobjects = np.array(self.DataSet.timeserieslist, dtype="double")
            gene_id = self.DataSet.gene_id
        else:
            try:
                alltimeseriesobjects = np.array(self.DataSet.tiu_timeserieslist, dtype="double")
                gene_id = self.DataSet.tiu_gene_id
            except ValueError:
                print("Please run iso.total_isoform_usage() to calculate total isoform usage before clustering.")
            
        func = self.CLUSTERING_ALGORITHMS[self.algorithm]       
        simfunc = self.PROTOTYPE[self.prototypefunction]

        ###todo###

        #combine replicates
        allrep = simfunc(alltimeseriesobjects, axis=0)
        

        if self.transform is not None:
            transfunc = self.TRANSFORMATION[self.transform] 
            allrep = transfunc(allrep)
            #if value returns infinite , set it as 0
            allrep[~np.isfinite(allrep)]=0

        self.allrep=allrep 

        def _cal_dist_mat(self):
            if self.metric == "soft_dtw":
                dist = cdist_soft_dtw(allrep)
                dist[np.isnan(dist)] = 0
            else:
                #compute euclidean/correlation 
                dist = pairwise_distances(allrep, metric=self.metric)
                dist[np.isnan(dist)] = 0 
            return dist

        dist = _cal_dist_mat(self)

        #compute inverse shortest distance 
        if self.composite:
            sp, gene_index = inverse_shortest_path(gene_id, self.BioNetwork)

            ixgrid = np.ix_(gene_index, gene_index)
            sp_final = sp[ixgrid]

            #composite metrics
            dist = (dist+sp_final)/2

        #distance matrix diagonal should be 0
        np.fill_diagonal(dist, 0) #return none


        bestsim = -np.inf
        if self.algorithm == "hierarchical" and self.linkage != "ward":
            cluster = func(affinity="precomputed", linkage=self.linkage, n_clusters=self.n_clusters).fit(dist)
            #sil = silhouette_score(dist, cluster.labels_, metric="precomputed")

        elif self.algorithm == "hierarchical" and self.linkage=="ward":
            cluster = func(affinity="euclidean", linkage=self.linkage, n_clusters=self.n_clusters).fit(allrep)
            #sil = silhouette_score(allrep, cluster.labels_, metric="euclidean")

        elif self.algorithm == "kmedoids":
            cluster = func(n_clusters=self.n_clusters).fit(dist)
            #sil =  silhouette_score(dist, cluster.labels_, metric="precomputed")
        
        elif self.algorithm == "kmeans" and self.metric == "soft_dtw":
            func = TimeSeriesKMeans
            cluster = func(n_clusters=self.n_clusters, metric="softdtw").fit(allrep)
            #sil = tssilhouette_score(allrep, cluster.labels_, metric="softdtw", n_jobs=5)

        else:
            cluster = func(metric=self.metric, **self.kwargs).fit(allrep)
            #sil = silhouette_score(dist, cluster.labels_, metric="precomputed")

        sil = silhouette_score(allrep, cluster.labels_, metric="euclidean")
        db = davies_bouldin_score(allrep, cluster.labels_)
        
        self._labels = cluster.labels_+1
        self.silhouette_index = sil
        self.davies_bouldin = db
        self._map_clusters(self._labels)
        self._prototype = {}
        for u,v in self.index_clusters.items():
            self._prototype.update({u: simfunc(allrep[v], axis=0)})

        
        self.DataSet.clusterobj = self
        print(f"clustering took {time.time()-t1}s. ")
        return self.genelist_clusters


    def _map_clusters(self, clusterlabel):
        if self.input_type == "expression":
            genelist = self.DataSet.gene_id
            symbol = self.DataSet.symbs
        else:
            genelist = self.DataSet.tiu_gene_id
            symbol = self.DataSet.tiu_symbs

        
        #map clusters to gene list dict
        self.genelist_clusters = {}
        self.index_clusters = {}
        self.symbs_clusters = {}
        for idx, cluster in enumerate(clusterlabel):
            if cluster in self.genelist_clusters.keys():
                self.genelist_clusters[int(cluster)].append(genelist[idx])
                self.index_clusters[int(cluster)].append(idx)
                self.symbs_clusters[int(cluster)].append(symbol[idx])
            else:
                self.genelist_clusters.update({int(cluster):[genelist[idx]]})
                self.index_clusters.update({int(cluster):[idx]})
                self.symbs_clusters.update({int(cluster):[symbol[idx]]})

        self._final_n_cluster = len(self.genelist_clusters)

    def combine_clusters(self, clusters_to_combine):
        update_genelist = defaultdict(list)
        update_index = defaultdict(list)
        update_symbs = defaultdict(list)
        for u in self.genelist_clusters.keys():
            if u in clusters_to_combine:
                update_genelist[clusters_to_combine[0]].append(self.genelist_clusters[u])
                update_index[clusters_to_combine[0]].append(self.index_clusters[u])
                update_symbs[clusters_to_combine[0]].append(self.symbs_clusters[u])
            else:
                update_genelist[u].append(self.genelist_clusters[u])
                update_index[u].append(self.index_clusters[u])
                update_symbs[u].append(self.symbs_clusters[u])

        for u in update_genelist.keys():
            update_genelist[u] = reduce(lambda a,y:a+y, update_genelist[u])
            update_index[u] = reduce(lambda a,y: a+y, update_index[u])
            update_symbs[u] = reduce(lambda a,y:a+y, update_symbs[u])

        self.genelist_clusters = update_genelist
        self.index_clusters = update_index
        self.symbs_clusters = update_symbs
        self._prototype = {}
        for u,v in self.index_clusters.items():
            self._prototype.update({u: self.PROTOTYPE[self.prototypefunction](self.allrep[v], axis=0)})




    # def calculate_pvalue(self, object_type = "clusters", n_permutations=1000, fitness_scores_two_sided = True):
    #     cal = compute_pvalues.basic_pvalue(testobject=self.DataSet, object_type = object_type, n_permutations = n_permutations, fitness_scores_two_sided = fitness_scores_two_sided)
    #     pval = cal.do_calculate()

    #     self.cluster_pvalues = np.array(pval)
        
    #     return np.array(pval)
        #fitness_scores.append()


import pandas as pd
import numpy as np
import time
import random
import warnings
import math
import itertools as it
from sklearn.metrics import adjusted_rand_score, jaccard_score, adjusted_mutual_info_score

from ..DataSet import DataSet
from ..BioNetwork import BioNetwork
from .._clustering.clusterobj import clusterObj
from .._clustering import similarities as sim


def jaccard(c,cnorm):
    intersect = len(set(c).intersection(set(cnorm)))
    if intersect == 0:
        return 0
    total = float(intersect) / len(set(c).union(set(cnorm)))
    return total

def AMIscore(clusterObj, groundtruth):
    '''
    groundtruth: dataframe with two columns names: cluster and object, which is the correct classification of the objects.
    '''
    #convert groundtruth to dict
    '''
    Compute the similarity measure between two clusterings
    '''
    clusters = clusterObj.symbs_clusters
    testdict = {}
    keys = groundtruth['cluster'].unique()
    for key in keys:
        testdict[key] = list(groundtruth[groundtruth['cluster']==key].object)

    true = []
    for c,o in testdict.items():
        for i in o:
            true.append(c)

    pred = []
    for obj in list(groundtruth.object):
        tmp = [pred.append(k) for k, v in clusters.items() for i in v if i == obj]

    return adjusted_mutual_info_score(true, pred)


def adjusted_rand_index(clusterObj, groundtruth):
    '''
    Compute the similarity measure between two clusterings
    '''
    clusters = clusterObj.symbs_clusters
    testdict = {}
    keys = groundtruth['cluster'].unique()
    for key in keys:
        testdict[key] = list(groundtruth[groundtruth['cluster']==key].object)

    true = []
    for c,o in testdict.items():
        for i in o:
            true.append(c)

    pred = []
    for obj in list(groundtruth.object):
        tmp = [pred.append(k) for k, v in clusters.items() for i in v if i == obj]

    return adjusted_rand_score(true, pred)


import sys
sys.path.insert(0, "../")
import networkx as nx
import pandas as pd
import community as community_louvain
import numpy as np

def create_slices(network_file, output_file_name, resolution=0.15):
    if type(network_file)==str:
        df = pd.read_csv(network_file, sep='\t', dtype=str)
        df.columns = ["node_1", "node_2"]
        G = nx.from_pandas_edgelist(df, 'node_1', 'node_2')
    else:
        G = network_file.networkz

    partition = community_louvain.best_partition(G, resolution=resolution, random_state=1)  # 0.1
    prt = {k: [] for k in np.arange(len(np.unique(list(partition.values()))))}
    for k, v in partition.items():
        prt[v].append(k)

    i = 0
    with open(output_file_name, 'w+') as f:
        f.write(f'# of cc after modularity optimization: {len(prt.keys())}\n')
        for k, v in prt.items():
            if len(v) >= 10:
                f.write(f'cc #{i}: n={len(v)}\n[{", ".join(v)}]\n')
                i += 1


def read_preprocessed_slices(file_path):
    modules = []

    with open(file_path, 'r') as f:
        line = f.readline()
        while line != "":
            line = f.readline()
            if line.startswith("cc"):
                modules.append(f.readline().strip()[1:-1].split(', '))

        return modules
